# coding: utf-8

"""
Simple ARC job manager. See http://www.nordugrid.org/arc and
http://www.nordugrid.org/documents/xrsl.pdf.
"""

__all__ = ["ARCJobManager", "ARCJobFileFactory"]


import os
import stat
import time
import re
import random
import subprocess

from law.config import Config
from law.job.base import BaseJobManager, BaseJobFileFactory, JobInputFile, DeprecatedInputFiles
from law.target.file import get_path
from law.util import interruptable_popen, make_list, make_unique, quote_cmd
from law.logger import get_logger


logger = get_logger(__name__)

_cfg = Config.instance()


class ARCJobManager(BaseJobManager):

    # chunking settings
    chunk_size_submit = _cfg.get_expanded_int("job", "arc_chunk_size_submit")
    chunk_size_cancel = _cfg.get_expanded_int("job", "arc_chunk_size_cancel")
    chunk_size_cleanup = _cfg.get_expanded_int("job", "arc_chunk_size_cleanup")
    chunk_size_query = _cfg.get_expanded_int("job", "arc_chunk_size_query")

    submission_job_id_cre = re.compile("^Job submitted with jobid: (.+)$")
    status_block_cre = re.compile(r"\s*([^:]+): (.*)\n")
    status_invalid_job_cre = re.compile("^.+: Job not found in job list: (.+)$")
    status_missing_job_cre = re.compile(
        "^.+: Job information not found in the information system: (.+)$")

    def __init__(self, job_list=None, ce=None, threads=1):
        super(ARCJobManager, self).__init__()

        self.job_list = job_list
        self.ce = ce
        self.threads = threads

    def submit(self, job_file, job_list=None, ce=None, retries=0, retry_delay=3, silent=False):
        # default arguments
        if job_list is None:
            job_list = self.job_list
        if ce is None:
            ce = self.ce

        # check arguments
        if not ce:
            raise ValueError("ce must not be empty")
        ce = make_list(ce)

        # arc supports multiple jobs to be submitted with a single arcsub call,
        # so job_file can be a sequence of files
        # when this is the case, we have to make the assumption that their input files are all
        # absolute, or they are relative but all in the same directory
        chunking = isinstance(job_file, (list, tuple))
        job_files = list(map(str, make_list(job_file)))
        job_file_dir = os.path.dirname(os.path.abspath(job_files[0]))
        job_file_names = [os.path.basename(jf) for jf in job_files]

        # define the actual submission in a loop to simplify retries
        while True:
            # build the command
            cmd = ["arcsub", "-c", random.choice(ce)]
            if job_list:
                cmd += ["-j", job_list]
            cmd += job_file_names
            cmd = quote_cmd(cmd)

            # run the command
            logger.debug("submit arc job(s) with command '{}'".format(cmd))
            code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
                stdout=subprocess.PIPE, cwd=job_file_dir)

            # in some cases, the return code is 0 but the ce did not respond valid job ids
            job_ids = []
            if code == 0:
                for line in out.strip().split("\n"):
                    m = self.submission_job_id_cre.match(line.strip())
                    if m:
                        job_id = m.group(1)
                        job_ids.append(job_id)

                if not job_ids:
                    code = 1
                    out = "cannot find job id(s) in output:\n{}".format(out)
                elif len(job_ids) != len(job_files):
                    raise Exception("number of job ids in output ({}) does not match number of "
                        "jobs to submit ({}) in output:\n{}".format(len(job_ids), len(job_files),
                        out))

            # retry or done?
            if code == 0:
                return job_ids if chunking else job_ids[0]

            logger.debug("submission of arc job(s) '{}' failed with code {}:\n{}".format(
                job_files, code, out))

            if retries > 0:
                retries -= 1
                time.sleep(retry_delay)
                continue

            if silent:
                return None

            raise Exception("submission of arc job(s) '{}' failed:\n{}".format(job_files, out))

    def cancel(self, job_id, job_list=None, silent=False):
        # default arguments
        if job_list is None:
            job_list = self.job_list

        chunking = isinstance(job_id, (list, tuple))
        job_ids = make_list(job_id)

        # build the command
        cmd = ["arckill"]
        if job_list:
            cmd += ["-j", job_list]
        cmd += job_ids
        cmd = quote_cmd(cmd)

        # run it
        logger.debug("cancel arc job(s) with command '{}'".format(cmd))
        code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE)

        # check success
        if code != 0 and not silent:
            # arc prints everything to stdout
            raise Exception("cancellation of arc job(s) '{}' failed with code {}:\n{}".format(
                job_id, code, out))

        return {job_id: None for job_id in job_ids} if chunking else None

    def cleanup(self, job_id, job_list=None, silent=False):
        # default arguments
        if job_list is None:
            job_list = self.job_list

        chunking = isinstance(job_id, (list, tuple))
        job_ids = make_list(job_id)

        # build the command
        cmd = ["arcclean"]
        if job_list:
            cmd += ["-j", job_list]
        cmd += job_ids
        cmd = quote_cmd(cmd)

        # run it
        logger.debug("cleanup arc job(s) with command '{}'".format(cmd))
        code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE)

        # check success
        if code != 0 and not silent:
            # arc prints everything to stdout
            raise Exception("cleanup of arc job(s) '{}' failed with code {}:\n{}".format(
                job_id, code, out))

        return {job_id: None for job_id in job_ids} if chunking else None

    def query(self, job_id, job_list=None, silent=False):
        # default arguments
        if job_list is None:
            job_list = self.job_list

        chunking = isinstance(job_id, (list, tuple))
        job_ids = make_list(job_id)

        # build the command
        cmd = ["arcstat"]
        if job_list:
            cmd += ["-j", job_list]
        cmd += job_ids
        cmd = quote_cmd(cmd)

        # run it
        logger.debug("query arc job(s) with command '{}'".format(cmd))
        code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        # handle errors
        if code != 0:
            if silent:
                return None
            else:
                # arc prints everything to stdout
                raise Exception("status query of arc job(s) '{}' failed with code {}:\n{}".format(
                    job_id, code, out))

        # parse the output and extract the status per job
        query_data = self.parse_query_output(out)

        # compare to the requested job ids and perform some checks
        for _job_id in job_ids:
            if _job_id not in query_data:
                if not chunking:
                    if silent:
                        return None
                    else:
                        raise Exception("arc job(s) '{}' not found in query response".format(
                            job_id))
                else:
                    query_data[_job_id] = self.job_status_dict(job_id=_job_id, status=self.FAILED,
                        error="job not found in query response")

        return query_data if chunking else query_data[job_id]

    @classmethod
    def parse_query_output(cls, out):
        query_data = {}

        # first, check for invalid and missing jobs
        for line in out.strip().split("\n"):
            line = line.strip()

            # invalid job?
            m = cls.status_invalid_job_cre.match(line)
            if m:
                job_id = m.group(1)
                query_data[job_id] = cls.job_status_dict(job_id=job_id, status=cls.FAILED, code=1,
                    error="job not found")
                continue

            # missing job? this means that the job is not yet present in the information system
            # so it can be considered pending
            m = cls.status_missing_job_cre.match(line)
            if m:
                job_id = m.group(1)
                query_data[job_id] = cls.job_status_dict(job_id=job_id, status=cls.PENDING)
                continue

        # retrieve actual job status information per block
        # remove the summary line and check if there is any valid job status
        out = out.split("\nStatus of ", 1)[0].strip()
        if "Job: " not in out:
            return query_data

        blocks = out.split("Job: ", 1)[1].strip().split("\nJob: ")
        for block in blocks:
            data = dict(cls.status_block_cre.findall("Job: {}\n".format(block)))
            if not data:
                continue

            # get the job id
            if "Job" not in data:
                continue
            job_id = data["Job"]

            # interpret data
            status = cls.map_status(data.get("State") or None)
            code = data.get("Exit Code") and int(data["Exit Code"])
            error = data.get("Job Error") or None

            # special cases
            if status == cls.FAILED and code in (0, None):
                code = 1

            # store it
            query_data[job_id] = cls.job_status_dict(job_id=job_id, status=status, code=code,
                error=error)

        return query_data

    @classmethod
    def map_status(cls, status):
        # see http://www.nordugrid.org/documents/arc-ui.pdf
        if status in ("Queuing", "Accepted", "Preparing", "Submitting"):
            return cls.PENDING
        elif status in ("Running", "Finishing"):
            return cls.RUNNING
        elif status in ("Finished",):
            return cls.FINISHED
        elif status in ("Failed", "Deleted"):
            return cls.FAILED
        else:
            return cls.FAILED


class ARCJobFileFactory(BaseJobFileFactory):

    config_attrs = BaseJobFileFactory.config_attrs + [
        "file_name", "command", "executable", "arguments", "input_files", "output_files",
        "postfix_output_files", "output_uri", "overwrite_output_files", "job_name", "log", "stdout",
        "stderr", "custom_content", "absolute_paths",
    ]

    def __init__(self, file_name="arc_job.xrsl", command=None, executable=None, arguments=None,
            input_files=None, output_files=None, postfix_output_files=True, output_uri=None,
            overwrite_output_files=True, job_name=None, log="log.txt", stdout="stdout.txt",
            stderr="stderr.txt", custom_content=None, absolute_paths=True, **kwargs):
        # get some default kwargs from the config
        cfg = Config.instance()
        if kwargs.get("dir") is None:
            kwargs["dir"] = cfg.get_expanded("job", cfg.find_option("job",
                "arc_job_file_dir", "job_file_dir"))
        if kwargs.get("mkdtemp") is None:
            kwargs["mkdtemp"] = cfg.get_expanded_bool("job", cfg.find_option("job",
                "arc_job_file_dir_mkdtemp", "job_file_dir_mkdtemp"))
        if kwargs.get("cleanup") is None:
            kwargs["cleanup"] = cfg.get_expanded_bool("job", cfg.find_option("job",
                "arc_job_file_dir_cleanup", "job_file_dir_cleanup"))

        super(ARCJobFileFactory, self).__init__(**kwargs)

        self.file_name = file_name
        self.command = command
        self.executable = executable
        self.arguments = arguments
        self.input_files = DeprecatedInputFiles(input_files or {})
        self.output_files = output_files or []
        self.postfix_output_files = postfix_output_files
        self.output_uri = output_uri
        self.overwrite_output_files = overwrite_output_files
        self.job_name = job_name
        self.log = log
        self.stdout = stdout
        self.stderr = stderr
        self.absolute_paths = absolute_paths
        self.custom_content = custom_content

    def create(self, postfix=None, **kwargs):
        # merge kwargs and instance attributes
        c = self.get_config(**kwargs)

        # some sanity checks
        if not c.file_name:
            raise ValueError("file_name must not be empty")
        if not c.command and not c.executable:
            raise ValueError("either command or executable must not be empty")

        # ensure that all log files are output files
        for attr in ["log", "stdout", "stderr", "custom_log_file"]:
            if c[attr] and c[attr] not in c.output_files:
                c.output_files.append(c[attr])

        # postfix certain output files
        c.output_files = list(map(str, c.output_files))
        if c.postfix_output_files:
            c.output_files = [self.postfix_output_file(path, postfix) for path in c.output_files]
            for attr in ["log", "stdout", "stderr", "custom_log_file"]:
                if c[attr]:
                    c[attr] = self.postfix_output_file(c[attr], postfix)

        # ensure that all input files are JobInputFile's
        c.input_files = {
            key: JobInputFile(f)
            for key, f in c.input_files.items()
        }

        # special case: remote input files must never be copied
        for f in c.input_files.values:
            if f.is_remote:
                f.copy = False

        # ensure that the executable is an input file, remember the key to access it
        if c.executable:
            executable_keys = [
                k
                for k, v in c.input_files.items()
                if get_path(v) == get_path(c.executable)
            ]
            if executable_keys:
                executable_key = executable_keys[0]
            else:
                executable_key = "executable_file"
                c.input_files[executable_key] = JobInputFile(c.executable)

        # prepare input files
        def prepare_input(f):
            if f.is_remote:
                return f.path
            # when not copied or forwarded, just return the absolute, original path
            abs_path = os.path.abspath(f.path)
            if not f.copy or f.forward:
                return abs_path
            # copy the file
            abs_path = self.provide_input(
                src=abs_path,
                postfix=postfix if f.postfix and not f.share else None,
                dir=c.dir,
                skip_existing=f.share,
            )
            return abs_path

        # absolute input paths
        for key, f in c.input_files.items():
            f.path_sub_abs = prepare_input(f)

        # input paths relative to the submission or initial dir
        # forwarded files are included but remote ones are skipped
        for key, f in c.input_files.items():
            if f.is_remote:
                continue
            f.path_sub_rel = (
                os.path.basename(f.path_sub_abs)
                if f.copy and not c.absolute_paths else
                f.path_sub_abs
            )

        # input paths as seen by the job, before and after potential rendering
        for key, f in c.input_files.items():
            f.path_job_pre_render = (
                f.path_sub_abs
                if f.is_remote else
                os.path.basename(f.path_sub_abs)
            )
            f.path_job_post_render = (
                os.path.basename(f.path_sub_abs)
                if f.render_job else
                f.path_sub_abs
            )

        # update files in render variables with version after potential rendering
        c.render_variables.update({
            key: f.path_job_post_render
            for key, f in c.input_files.items()
        })

        # add space separated input files before potential rendering to render variables
        c.render_variables["input_files"] = " ".join(
            f.path_job_pre_render
            for f in c.input_files.values()
        )

        # add space separated list of input files for rendering
        c.render_variables["input_files_render"] = " ".join(
            f.path_job_pre_render
            for f in c.input_files.values()
            if f.render_job
        )

        # add the custom log file to render variables
        if c.custom_log_file:
            c.render_variables["log_file"] = c.custom_log_file

        # add the file postfix to render variables
        if postfix and "file_postfix" not in c.render_variables:
            c.render_variables["file_postfix"] = postfix

        # add output_uri to render variables
        if c.output_uri and "output_uri" not in c.render_variables:
            c.render_variables["output_uri"] = c.output_uri

        # linearize render variables
        render_variables = self.linearize_render_variables(c.render_variables)

        # prepare the job file
        job_file = self.postfix_input_file(os.path.join(c.dir, str(c.file_name)), postfix)

        # render copied, non-remote input files
        for key, f in c.input_files.items():
            if not f.copy or f.is_remote or not f.render_local:
                continue
            self.render_file(
                f.path_sub_abs,
                f.path_sub_abs,
                render_variables,
                postfix=postfix if f.postfix else None,
            )

        # create arc-style input file pairs
        input_file_pairs = [
            (os.path.basename(f.path_sub_abs), "" if f.copy else f.path_sub_abs)
            for key, f in c.input_files.items()
        ]

        # prepare the executable when given
        if c.executable:
            c.executable = get_path(c.input_files[executable_key].path_job_post_render)
            # make the file executable for the user and group
            path = os.path.join(c.dir, os.path.basename(c.executable))
            if os.path.exists(path):
                os.chmod(path, os.stat(path).st_mode | stat.S_IXUSR | stat.S_IXGRP)

        # ensure a correct format of output files
        def prepare_output(path):
            # consider strings to be the filename and when output_uri is set, use it
            # as the URL, otherwise it's also empty
            if c.postfix_output_files:
                path = self.postfix_output_file(path, postfix)
            dst = os.path.join(c.output_uri, os.path.basename(path)) if c.output_uri else ""
            opts = "overwrite=yes" if c.overwrite_output_files else None
            return (path, dst, opts) if opts else (path, dst)

        c.output_files = list(map(prepare_output, c.output_files))

        # job file content
        content = []
        if c.command:
            cmd = quote_cmd(c.command) if isinstance(c.command, (list, tuple)) else c.command
            content.append(("executable", cmd))
        elif c.executable:
            content.append(("executable", c.executable))
        if c.arguments:
            args = quote_cmd(c.arguments) if isinstance(c.arguments, (list, tuple)) else c.arguments
            content.append(("arguments", args))
        if c.job_name:
            content.append(("jobName", c.job_name))
        if c.input_files:
            content.append(("inputFiles", make_unique(input_file_pairs)))
        if c.output_files:
            content.append(("outputFiles", make_unique(c.output_files)))
        if c.log:
            content.append(("gmlog", c.log))
        if c.stdout:
            content.append(("stdout", c.stdout))
        if c.stderr:
            content.append(("stderr", c.stderr))

        # add custom content
        if c.custom_content:
            content += c.custom_content

        # write the job file
        with open(job_file, "w") as f:
            f.write("&\n")
            for key, value in content:
                line = self.create_line(key, value)
                f.write(line + "\n")

        logger.debug("created arc job file at '{}'".format(job_file))

        return job_file, c

    @classmethod
    def create_line(cls, key, value):
        def flat_value(value):
            if isinstance(value, list):
                return " ".join(flat_value(v) for v in value)
            if isinstance(value, tuple):
                return "({})".format(" ".join(flat_value(v) for v in value))
            else:
                return "\"{}\"".format(value)
        return "({} = {})".format(key, flat_value(value))



# coding: utf-8

"""
Local workflow implementation.
"""

__all__ = ["LocalWorkflow"]

from collections.abc import Generator

import luigi

from law import luigi_version_info
from law.workflow.base import BaseWorkflow, BaseWorkflowProxy
from law.target.collection import SiblingFileCollectionBase
from law.logger import get_logger
from law.util import DotDict


logger = get_logger(__name__)


class LocalWorkflowProxy(BaseWorkflowProxy):
    """
    Workflow proxy class for the local workflow implementation. The workflow type is ``"local"``.
    """

    workflow_type = "local"

    def __init__(self, *args, **kwargs):
        super(LocalWorkflowProxy, self).__init__(*args, **kwargs)

        self._local_workflow_has_yielded = False

    def requires(self):
        reqs = super(LocalWorkflowProxy, self).requires()

        local_reqs = self.task.local_workflow_requires()
        if local_reqs:
            reqs.update(local_reqs)

        # when local_workflow_require_branches is True, add all branch tasks as dependencies
        if self.task.local_workflow_require_branches:
            reqs["branches"] = self.task.get_branch_tasks()

        return reqs

    def run(self):
        """
        When *local_workflow_require_branches* of the task was set to *False*, starts all branch
        tasks via dynamic dependencies by yielding them in a list, or simply does nothing otherwise.
        """
        pre_run_gen = self.task.local_workflow_pre_run()
        if isinstance(pre_run_gen, Generator):
            yield pre_run_gen

        super(LocalWorkflowProxy, self).run()

        if not self.task.local_workflow_require_branches and not self._local_workflow_has_yielded:
            self._local_workflow_has_yielded = True

            # use branch tasks as requirements
            branch_tasks = self.task.get_branch_tasks()
            reqs = list(branch_tasks.values())

            # helper to get the output collection
            get_col = lambda: self.get_cached_output().get("collection")

            # wrap into DynamicRequirements when available, otherwise just yield the list
            if luigi_version_info[:3] >= (3, 1, 2):
                # in case the workflows creates a sibling file collection, per-branch completion
                # checks are possible in advance and can be stored in luigi's completion cache
                def custom_complete(complete_fn):
                    # get the cache (stored as a specified keyword of a partial'ed function)
                    cache = getattr(complete_fn, "keywords", {}).get("completion_cache")
                    if cache is None:
                        if complete_fn(self):
                            return True
                        # show a warning for large workflows that use sibling file collections and
                        # that could profit from the cache_task_completion feature
                        if len(reqs) >= 100 and isinstance(get_col(), SiblingFileCollectionBase):
                            url = "https://luigi.readthedocs.io/en/stable/configuration.html#worker"
                            logger.warning_once(
                                "cache_task_completion_hint",
                                "detected SiblingFileCollection for LocalWorkflow with {} branches "
                                "whose completness checks will be performed manually by luigi; "
                                "consider enabling luigi's cache_task_completion feature to speed "
                                "up these checks; fore more info, see {}".format(len(reqs), url),
                            )
                        return False

                    # the output collection must be a sibling file collection
                    col = get_col()
                    if not isinstance(col, SiblingFileCollectionBase):
                        return complete_fn(self)

                    # get existing branches and populate the cache with completeness states
                    existing_branches = set(col.count(keys=True)[1])
                    for b, task in branch_tasks.items():
                        cache[task.task_id] = b in existing_branches

                    # finally, evaluate the normal completeness check on the workflow
                    return complete_fn(self)

                yield luigi.DynamicRequirements(reqs, custom_complete)

            else:
                # old, possibly slow behavior
                yield reqs


class LocalWorkflow(BaseWorkflow):
    """
    Local workflow implementation. The workflow type is ``"local"``. There are two ways how a local
    workflow starts its branch tasks. See the :py:attr:`local_workflow_require_branches` attribute
    for more information.

    Since local workflows trigger their branch tasks via requirements or dynamic dependencies, their
    run methods do not support decorators. See :py:attr:`BaseWorkflow.workflow_run_decorators` for
    more info.

    .. py:classattribute:: workflow_proxy_cls

        type: :py:class:`BaseWorkflowProxy`

        Reference to the :py:class:`LocalWorkflowProxy` class.

    .. py:classattribute:: local_workflow_require_branches

        type: bool

        When *True*, the workflow will require its branch tasks within
        :py:meth:`LocalWorkflowProxy.requires` so that the execution of the workflow indirectly
        starts all branch tasks. When *False*, the workflow uses dynamic dependencies by yielding
        its branch tasks within its own run method.
    """

    workflow_proxy_cls = LocalWorkflowProxy

    local_workflow_require_branches = False

    exclude_index = True

    def local_workflow_requires(self):
        return DotDict()

    def local_workflow_pre_run(self):
        return



# coding: utf-8

"""
Workflow and workflow proxy base class definitions.
"""

__all__ = [
    "BaseWorkflow", "WorkflowParameter", "workflow_property", "dynamic_workflow_condition",
    "DynamicWorkflowCondition",
]


import re
import copy
import functools
import itertools
import inspect
from collections import OrderedDict, defaultdict
from abc import abstractmethod

import luigi
import six

from law.task.base import Register
from law.task.proxy import ProxyTask, ProxyAttributeTask
from law.target.collection import TargetCollection
from law.target.local import LocalFileTarget
from law.parameter import NO_STR, MultiRangeParameter, CSVParameter
from law.util import (
    no_value, make_list, make_set, iter_chunks, range_expand, range_join, create_hash,
    is_classmethod, DotDict,
)
from law.logger import get_logger


logger = get_logger(__name__)


class BaseWorkflowProxy(ProxyTask):
    """
    Base class of all workflow proxies.

    .. py:classattribute:: workflow_type

        type: string

        The named type of the workflow. This attribute refers to the value of the ``--workflow``
        parameter on the command line to select a particular workflow.

    .. py:attribute:: task

        type: Task

        Reference to the actual *workflow* task.
    """

    workflow_type = None

    add_workflow_run_decorators = True

    def __init__(self, *args, **kwargs):
        super(BaseWorkflowProxy, self).__init__(*args, **kwargs)

        # find decorators for this proxy's run method that can be configured on the actual task
        if self.add_workflow_run_decorators:
            for prefix in [self.workflow_type + "_", ""]:
                attr = "{}workflow_run_decorators".format(prefix)
                decorators = getattr(self.task, attr, None)
                if decorators is not None:
                    # found decorators, so unbound, decorate and re-bound
                    run_func = self.run.__func__
                    for decorator in decorators:
                        run_func = decorator(run_func)
                    self.run = run_func.__get__(self, self.__class__)
                    break

        self._workflow_has_reset_branch_map = False

        # cached outputs
        self._cached_output = None

    def _get_task_attribute(self, name, fallback=False):
        """
        Return an attribute of the actual task named ``<workflow_type>_<name>``. When the attribute
        does not exist and *fallback* is *True*, try to return the task attribute simply named
        *name*. *name* can also be a sequence of strings that are check in the given order. In this
        case, the *fallback* option is not considered.

        Eventually, if no matching attribute is found, an AttributeError is raised.
        """
        if isinstance(name, (list, tuple)):
            attributes = name
        else:
            attributes = ["{}_{}".format(self.workflow_type, name)]
            if fallback:
                attributes.append(name)

        for attr in attributes:
            value = getattr(self.task, attr, no_value)
            if value != no_value:
                return value

        raise AttributeError("'{!r}' object has none of the requested attribute(s) {}".format(
            self, ",".join(map(str, attributes)),
        ))

    def complete(self):
        """
        Custom completion check that invokes the task's *workflow_complete* method and if it returns
        anything else than *NotImplemented* returns the value, or just does the default completion
        check otherwise.
        """
        complete = self.task.workflow_complete()
        if complete is not NotImplemented:
            return complete

        return super(BaseWorkflowProxy, self).complete()

    def requires(self):
        """
        Returns the default workflow requirements in an ordered dictionary, which is updated with
        the return value of the task's *workflow_requires* method.
        """
        reqs = DotDict()
        workflow_reqs = self.task.workflow_requires()
        if workflow_reqs:
            reqs.update(workflow_reqs)
        return reqs

    def output(self):
        """
        Returns the default workflow outputs in an ordered dictionary. At the moment this is just
        the collection of outputs of the branch tasks, stored with the key ``"collection"``.
        """
        # get all targets
        targets = luigi.task.getpaths(self.task.get_branch_tasks())

        # determine the collection container clas
        cls = TargetCollection
        if self.task.output_collection_cls and targets:
            cls = self.task.output_collection_cls

        # create the collection
        collection = cls(targets, threshold=self.threshold(len(targets)))
        return DotDict([("collection", collection)])

    def get_cached_output(self, update=False):
        """
        If already cached, returns the previously computed output, and otherwise computes it via
        :py:meth:`output` and caches it for subsequent calls, if :py:attr:`cache_brach_map` of the
        task is *True*.
        """
        # invalidate cache
        if update:
            self._cached_output = None

        # return from cache if present
        if self._cached_output is not None:
            return self._cached_output

        # get output and cache it if possible
        output = self.output()
        if self.task.cache_branch_map:
            self._cached_output = output

        return output

    def threshold(self, n=None):
        """
        Returns the threshold number of tasks that need to be complete in order to consider the
        workflow as being complete itself. This takes into account the
        :py:attr:`law.BaseWorkflow.acceptance` parameter of the workflow. The threshold is passed
        to the :py:class:`law.TargetCollection` (or :py:class:`law.SiblingFileCollection`) within
        :py:meth:`output`. By default, the maximum number of tasks is taken from the length of the
        branch map. For performance purposes, you can set this value, *n*, directly.
        """
        if n is None:
            n = len(self.task.get_branch_map())

        acceptance = self.task.acceptance
        return (acceptance * n) if acceptance <= 1 else acceptance

    def run(self):
        """
        Default run implementation that resets the branch map once if requested.
        """
        if self.task.reset_branch_map_before_run and not self._workflow_has_reset_branch_map:
            self._workflow_has_reset_branch_map = True

            # reset cached branch map, branch tasks and boundaries
            self.task._branch_map = None
            self.task._branch_tasks = None
            self.task.branches = self.task._initial_branches


def workflow_property(func=None, attr=None, setter=True, cache=False, empty_value=no_value):
    """
    Decorator to declare an attribute that is stored only on a workflow and optionally cached for
    subsequent calls. Therefore, the decorated method is expected to (lazily) provide the value to
    cache if enabled. When the value is equal to *empty_value*, it is not cached and the next access
    to the property will invoke the decorated method again. The resulting value is stored as either
    ``_workflow_<func.__name__>`` or ``_workflow_cached_<func.__name__>`` on the workflow. By
    default, a setter is provded to overwrite the the attribute. Set *setter* to *False* to disable
    this feature. Example:

    .. code-block:: python

        class MyTask(Workflow):

            @workflow_property
            def common_data(self):
                # this method is always called with *self* being the *workflow*
                return some_demanding_computation()

            @workflow_property(attr="my_own_property", setter=False, cache=True)
            def common_data2(self):
                return some_other_computation()
    """
    def decorator(func):
        _attr = attr or "_workflow_{}{}".format("cached_" if cache else "", func.__name__)

        @functools.wraps(func)
        def getter(self):
            wf = self.as_workflow()
            if getattr(wf, _attr, empty_value) == empty_value or not cache:
                setattr(wf, _attr, func(wf))
            return getattr(wf, _attr)

        _setter = None
        if setter:
            def _setter(self, value):  # noqa: F811
                wf = self.as_workflow()
                setattr(wf, _attr, value)

            _setter.__name__ = func.__name__

        return property(fget=getter, fset=_setter)

    return decorator if func is None else decorator(func)


class WorkflowParameter(CSVParameter):

    def __init__(self, *args, **kwargs):
        # force an empty default value, disable single values being wrapped by tuples, and declare
        # the parameter as insignificant as they only act as a convenient branch lookup interface
        kwargs["default"] = no_value
        kwargs["force_tuple"] = False
        kwargs["significant"] = False

        super(WorkflowParameter, self).__init__(*args, **kwargs)

        # linearize the default
        self._default = no_value

    def parse(self, inp):
        """"""
        if inp in (None, NO_STR, no_value):
            return no_value

        return super(WorkflowParameter, self).parse(inp)

    def serialize(self, value):
        """"""
        if value in (None, no_value):
            return ""

        return super(WorkflowParameter, self).serialize(value)


def dynamic_workflow_condition(
    condition_fn=None,
    create_branch_map_fn=None,
    requires_fn=None,
    requires_eager_fn=None,
    output_fn=None,
    condition_as_workflow=False,
    cache_met_condition=True,
):
    """
    Decorator factory that is meant to wrap a workflow methods that defines a dynamic workflow
    condition, returning a :py:class:`DynamicWorkflowCondition` instance.
    """
    def decorator(condition_fn):
        return DynamicWorkflowCondition(
            condition_fn=condition_fn,
            create_branch_map_fn=create_branch_map_fn,
            requires_fn=requires_fn,
            requires_eager_fn=requires_eager_fn,
            output_fn=output_fn,
            condition_as_workflow=condition_as_workflow,
            cache_met_condition=cache_met_condition,
        )

    return decorator if condition_fn is None else decorator(condition_fn)


class DynamicWorkflowCondition(object):
    """
    Container for a workflow method that defines whether the branch map can be dynamically
    constructed or whether a placeholder should be used until the condition is met. Similar to
    Python's ``property``, instances of this class provide additional attributes for decorating
    other methods that usually depend on the branch map, such as branch requirements or outputs.

    It is recommended to use the :py:func:`dynamic_workflow_condition` decorator (factory).
    Example:

    .. code-block:: python

        class MyWorkflow(law.LocalWorkflow):

            def workflow_requires(self):
                # define requirements for the full workflow to start
                reqs = super().workflow_requires()
                reqs["files"] = OtherTask.req(self)
                return reqs

            @law.dynamic_workflow_condition
            def workflow_condition(self):
                # declare that the branch map can be built if the workflow requirement exists
                # note: self.input() refers to the outputs of tasks defined in workflow_requires()
                return self.input()["files"].exists()

            @workflow_condition.create_branch_map
            def create_branch_map(self):
                # let's assume that OtherTask produces a json file containing a list of objects
                # that _this_ workflows iterates over, so we can simply return this list here
                return self.input()["files"].load(formatter="json")

            def requires(self):
                # branch-level requirement
                # note: this is not really necessary, since the branch requirements are only
                # evaluated _after_ a branch map is built, so OtherTask must have been completed
                return OtherTask.req(self)

            @workflow_condition.output
            def output(self):
                # define the output
                return law.LocalFileTarget("file_{}.txt".format(self.branch))

            def run(self):
                # trivial run implementation
                self.output().touch()

    The condition is defined by ``workflow_condition`` which is decorated by *this* object. Once it
    is met, the branch map is fully created and cached (as usual) for subsequent calls.

    In addition, both ``create_branch_map()`` and ``output()`` are decorated with corresponding
    attributes of the initially decorated object. As a result, both methods will return placeholder
    objects as long as the condition is not met - the branch map will be considered empty and the
    output will refer to a temporary placeholder target that is never created. Note that two
    additional decorators exist for handling dynaminc branch-level requirements. ``requires()`` is
    meant to decorator the actual branch requirement and is evaluated once the condition is met.
    ``requires_eager()`` is evaluated if the condition is _not_ met and is meant to provide eager
    requirements that do not depend on the condition.

    As a consequence, the amended workflow is fully dynamic with its exact shape potentially
    depending heavily on conditions that are only known at runtime.

    Internally, the condition is evaluated by the calling task which is usually a workflow, but it
    can also be one of its branch tasks if, for instance, sandboxing is involved. Set
    *condition_as_workflow* to *True* to ensure that the condition is always evaluated by the
    workflow itself.

    In case the ``workflow_condition`` involves a costly computation, it is recommended to cache
    evluation of the condition by setting *cache_met_condition* argument to *True* or a string
    denoting the task instance attribute where the met condition is stored. In the first case,
    the attribute defaults to ``_dynamic_workflow_condition_met``.
    """

    _decorator_result = object()

    def __init__(
        self,
        condition_fn,
        create_branch_map_fn=None,
        requires_fn=None,
        requires_eager_fn=None,
        output_fn=None,
        condition_as_workflow=False,
        cache_met_condition=True,
    ):
        super(DynamicWorkflowCondition, self).__init__()

        # attributes
        self._condition_fn = condition_fn
        self._create_branch_map_fn = create_branch_map_fn
        self._requires_fn = requires_fn
        self._requires_eager_fn = requires_eager_fn
        self._output_fn = output_fn
        self.condition_as_workflow = condition_as_workflow
        self.cache_met_condition = cache_met_condition
        if self.cache_met_condition and not isinstance(cache_met_condition, str):
            self.cache_met_condition = "_dynamic_workflow_condition_met"

    def _wrap_condition_fn(self):
        if self._condition_fn is None:
            return None

        @functools.wraps(self._condition_fn)
        def condition(inst, *args, **kwargs):
            # when caching, and the condition is already met, return the cached value
            if self.cache_met_condition and getattr(inst, self.cache_met_condition, False):
                return getattr(inst, self.cache_met_condition)

            # evaluate the condition
            task = inst.as_workflow() if self.condition_as_workflow else inst
            is_met = self._condition_fn(task, *args, **kwargs)

            # write to cache if requested
            if self.cache_met_condition and is_met:
                setattr(inst, self.cache_met_condition, is_met)

            return is_met

        return condition

    def create_branch_map(self, create_branch_map_fn):
        # store the function
        self._create_branch_map_fn = create_branch_map_fn

        return self._decorator_result

    def _wrap_create_branch_map(self, bound_condition_fn):
        if self._create_branch_map_fn is None:
            return None

        @functools.wraps(self._create_branch_map_fn)
        def create_branch_map(inst, *args, **kwargs):
            if not bound_condition_fn():
                return [None]

            # enable branch map caching since the condition is met
            inst.cache_branch_map = True

            return self._create_branch_map_fn(inst, *args, **kwargs)

        return create_branch_map

    def requires(self, requires_fn):
        # store the function
        self._requires_fn = requires_fn

        return self._decorator_result

    def requires_eager(self, requires_eager_fn):
        # store the function
        self._requires_eager_fn = requires_eager_fn

        return self._decorator_result

    def _wrap_requires(self, bound_condition_fn):
        if self._requires_fn is None:
            return None

        @functools.wraps(self._requires_fn)
        def requires(inst, *args, **kwargs):
            if not bound_condition_fn():
                # eager requirements if present
                if self._requires_eager_fn is not None:
                    return self._requires_eager_fn(inst, *args, **kwargs)
                return []

            # enable branch map caching since the condition is met
            inst.cache_branch_map = True

            return self._requires_fn(inst, *args, **kwargs)

        return requires

    def _wrap_requires_eager(self, bound_condition_fn):
        if self._requires_eager_fn is None:
            return None

        @functools.wraps(self._requires_eager_fn)
        def requires_eager(inst, *args, **kwargs):
            # just forward the call
            return self._requires_eager_fn(inst, *args, **kwargs)

        return requires_eager

    def output(self, output_fn):
        # store the function
        self._output_fn = output_fn

        return self._decorator_result

    def _wrap_output(self, bound_condition_fn):
        if self._output_fn is None:
            return None

        @functools.wraps(self._output_fn)
        def output(inst, *args, **kwargs):
            if not bound_condition_fn():
                return LocalFileTarget(is_tmp="DYNAMIC_WORKFLOW_PLACEHOLDER")

            # enable branch map caching since the condition is met
            inst.cache_branch_map = True

            return self._output_fn(inst, *args, **kwargs)

        return output

    def _iter_wrappers(self, bound_condition_fn):
        if self._create_branch_map_fn is not None:
            yield "create_branch_map", self._wrap_create_branch_map(bound_condition_fn)

        if self._requires_fn is not None:
            yield "requires", self._wrap_requires(bound_condition_fn)

        if self._requires_eager_fn is not None:
            yield "requires_eager", self._wrap_requires_eager(bound_condition_fn)

        if self._output_fn is not None:
            yield "output", self._wrap_output(bound_condition_fn)

    def copy(self):
        return copy.deepcopy(self)


class WorkflowRegister(Register):

    def __new__(metacls, name, bases, classdict):
        # handle dynamic workflow conditions
        condition_attr = metacls.check_dynamic_workflow_conditions(name, classdict)
        if condition_attr:
            # store the attribute when found and disable the branch map caching by default
            classdict["_condition_attr"] = condition_attr
            classdict.setdefault("cache_branch_map_default", False)

        # store a flag on the created class whether it defined a new workflow_proxy_cls
        # this flag will define the classes in the mro to consider for instantiating the proxy
        classdict["_defined_workflow_proxy"] = "workflow_proxy_cls" in classdict

        # create and return the class
        return super(WorkflowRegister, metacls).__new__(metacls, name, bases, classdict)

    @classmethod
    def check_dynamic_workflow_conditions(metacls, name, classdict):
        # check that only one condition is present in classdict
        condition_attr = None
        for attr, value in classdict.items():
            if not isinstance(value, DynamicWorkflowCondition):
                continue
            if condition_attr:
                raise Exception(
                    "class '{}' defined with more than one DynamicWorkflowCondition, found "
                    "'{}' after previously registered '{}'".format(name, attr, condition_attr),
                )
            condition_attr = attr

        return condition_attr


class BaseWorkflow(six.with_metaclass(WorkflowRegister, ProxyAttributeTask)):
    """
    Base class of all workflows.

    .. py:classattribute:: workflow

        type: :py:class:`luigi.Parameter`

        Workflow type that refers to the workflow proxy implementation at instantiation / execution
        time. Empty default value.

    .. py:classattribute:: acceptance

        type: :py:class:`luigi.FloatParameter`

        Number of complete tasks to consider the workflow successful. Values larger than one are
        interpreted as absolute numbers, and as fractions otherwise. Defaults to *1.0*.

    .. py:classattribute:: tolerance

        type: :py:class:`luigi.FloatParameter`

        Number of failed tasks to still consider the workflow successful. Values larger than one are
        interpreted as absolute numbers, and as fractions otherwise. Defaults to *0.0*.

    .. py:classattribute:: branch

        type: :py:class:`luigi.IntParameter`

        The branch number to run this task for. *-1* means that this task is the actual *workflow*,
        rather than a *branch* task. Defaults to *-1*.

    .. py:classattribute:: branches

        type: :py:class:`law.MultiRangeParameter`

        Explicit list of branches or branch ranges to process. Empty default value.

    .. py:classattribute:: workflow_proxy_cls

        type: :py:class:`BaseWorkflowProxy`

        Reference to the workflow proxy class associated to this workflow.

    .. py:classattribute:: output_collection_cls

        type: :py:class:`law.TargetCollection`

        Configurable target collection class to use, such as
        :py:class:`target.collection.TargetCollection`, :py:class:`target.collection.FileCollection`
        or :py:class:`target.collection.SiblingFileCollection`.

    .. py:classattribute:: force_contiguous_branches

        type: bool

        Flag that denotes if this workflow is forced to use contiguous branch numbers, starting from
        0. If *False*, an exception is raised otherwise.

    .. py:classattribute:: reset_branch_map_before_run

        type: bool

        Flag that denotes whether the branch map should be recreated from scratch before the run
        method of the underlying workflow proxy is called.

    .. py:classattribute:: create_branch_map_before_repr

        type: bool

        Flag that denotes whether the branch map should be created (if not already done) before the
        task representation is created via :py:meth:`repr`.

    .. py:classattribute:: cache_workflow_requirements

        type: bool

        Whether workflow requirements should be evaluated only cached and cached afterwards in the
        :py:attr:`_cached_workflow_requirements` attribute. Defaults to *False*.

    .. py:classattribute:: cache_branch_map_default

        type: bool

        The initial default value of the :py:attr:`cache_branch_map` attribute that decides whether
        the branch map be created only once and then cached in the :py:attr:`_branch_map` attribute.
        Defaults to *True*.

    .. py:classattribute:: workflow_run_decorators

        type: sequence, None

        Sequence of decorator functions that will be conveniently used to decorate the workflow
        proxy's run method. This way, there is no need to subclass and reset the
        :py:attr:`workflow_proxy_cls` just to add a decorator. The value is *None* by default.

    .. py:attribute:: workflow_cls

        type: :py:class:`law.Register`

        Reference to the class of the realized workflow. This is especially helpful in case your
        derived class inherits from multiple workflows.

    .. py:attribute:: workflow_proxy

        type: :py:class:`BaseWorkflowProxy`

        Reference to the underlying workflow proxy instance.

    .. py:attribute:: branch_map

        type: dict (read-only)

        Shorthand for :py:meth:`get_branch_map`.

    .. py:attribute:: branch_data

        type: any (read-only)

        Shorthand for ``self.branch_map[self.branch]``.
    """

    workflow = luigi.Parameter(
        default=NO_STR,
        description="the type of the workflow to use; uses the first workflow type in the MRO when "
        "empty; default: empty",
    )
    effective_workflow = luigi.Parameter(
        default=NO_STR,
        description="do not set manually",
    )
    acceptance = luigi.FloatParameter(
        default=1.0,
        significant=False,
        description="number of finished tasks to consider the task successful; relative fraction "
        "(<= 1) or absolute value (> 1); default: 1.0",
    )
    tolerance = luigi.FloatParameter(
        default=0.0,
        significant=False,
        description="number of failed tasks to still consider the task successful; relative "
        "fraction (<= 1) or absolute value (> 1); default: 0.0",
    )
    pilot = luigi.BoolParameter(
        default=False,
        significant=False,
        description="disable certain configurable requirements of the workflow to let branch tasks "
        "resolve requirements on their own; default: False",
    )
    branch = luigi.IntParameter(
        default=-1,
        description="the branch number/index to run this task for; -1 means this task is the "
        "workflow; default: -1",
    )
    branches = MultiRangeParameter(
        default=(),
        require_start=False,
        require_end=False,
        single_value=True,
        description="comma-separated list of branches to select; each value can have the format "
        "'start:end' (end not included as per Python) to support range syntax; default: empty",
    )

    # caches
    _cls_branch_map_cache = {}

    # configuration members
    workflow_proxy_cls = BaseWorkflowProxy
    output_collection_cls = None
    force_contiguous_branches = False
    reset_branch_map_before_run = False
    create_branch_map_before_repr = False
    cache_workflow_requirements = False
    cache_branch_map_default = True
    passthrough_requested_workflow = True
    workflow_run_decorators = None

    # skip from indexing
    exclude_index = True

    # parameter exclusions
    exclude_params_req = {"effective_workflow"}
    exclude_params_index = {"effective_workflow"}
    exclude_params_repr = {"workflow"}
    exclude_params_branch = {"acceptance", "tolerance", "pilot", "branches"}
    exclude_params_workflow = {"branch"}

    def __new__(cls, *args, **kwargs):
        inst = super(BaseWorkflow, cls).__new__(cls)

        # bind wrappers present in the optional condition object
        condition_attr = getattr(cls, "_condition_attr", None)
        if condition_attr:
            condition = getattr(inst, condition_attr, None)
            if isinstance(condition, DynamicWorkflowCondition):
                # bind the condition method itself
                bound_condition_fn = condition._wrap_condition_fn().__get__(inst)
                setattr(inst, condition_attr, bound_condition_fn)

                # store the condition object itself
                setattr(inst, condition_attr + "_obj", condition)

                # bind wrapped methods that currently correspond to placeholders
                for attr, wrapper in condition._iter_wrappers(bound_condition_fn):
                    if getattr(inst, attr, None) != DynamicWorkflowCondition._decorator_result:
                        continue
                    setattr(inst, attr, wrapper.__get__(inst))

        return inst

    @classmethod
    def modify_param_values(cls, params):
        params = super(BaseWorkflow, cls).modify_param_values(params)

        # determine the default workflow type when not set
        if params.get("workflow") in [None, NO_STR]:
            params["workflow"] = cls.find_workflow_cls().workflow_proxy_cls.workflow_type

        # set the effective workflow parameter based on the actual resolution
        workflow_cls = cls.find_workflow_cls(
            name=params["workflow"],
            fallback_to_first=cls.passthrough_requested_workflow,
        )
        params["effective_workflow"] = workflow_cls.workflow_proxy_cls.workflow_type

        # resolve workflow parameters
        params = cls._resolve_workflow_parameters(params)

        return params

    @classmethod
    def _resolve_workflow_parameters(cls, params):
        """
        Handles the translation from workflow parameters to branch values, updating *params*
        in-place.
        """
        workflow_params = [
            (name, param, params.get(name, no_value))
            for name, param in cls.get_params()
            if isinstance(param, WorkflowParameter)
        ]

        # nothing to do when the task does not use workflow parameters
        if not workflow_params:
            return params

        # helper for error messages
        cjoin = lambda seq: ",".join(map(str, seq))
        wparams_repr = lambda: cjoin(map("{0[0]}={0[2]}".format, workflow_params))

        # when there are any workflow parameters, create_branch_map must be a classmethod since
        # there is no way of accessing this map before instantiation
        if not is_classmethod(cls.create_branch_map, cls):
            raise Exception(
                "{}.create_branch_map must be a classmethod accepting a single parameter (dict "
                "of parameter names and values) in case workflows use WorkflowParameter "
                "objects in order to perform branch value lookups prior to any task "
                "instantiation; found workflow parameter(s) {}".format(
                    cls.__name__, wparams_repr(),
                ),
            )

        # helper to extract an entry from branch data (usually a dict)
        def get_branch_value(branch, branch_data, key):
            if isinstance(branch_data, dict):
                if key in branch_data:
                    return branch_data[key]
            elif getattr(branch_data, key, no_value) != no_value:
                return getattr(branch_data, key)
            raise AttributeError(
                "attribute or item '{}' unknown to branch data at branch {}: {}".format(
                    key, branch, branch_data,
                ),
            )

        # get the branch map, potentially from a cache
        try:
            # create a hash of all significant parameters to store the map
            h = hash((cls.task_family, tuple(params.items())))
        except TypeError:
            # some parameter is not hashable
            h = None

        # recreate the maps if needed
        branch_map, branch_map_reversed = (
            cls._cls_branch_map_cache[h]
            if h and h in cls._cls_branch_map_cache
            else (None, None)
        )
        if branch_map is None:
            # get the map and sanitize it
            branch_map = cls.create_branch_map(params)
            branch_map = cls._sanitize_branch_map(branch_map, cls.force_contiguous_branches)
            # create the reversed map, using workflow parameter value tuples as keys
            branch_map_reversed = OrderedDict()
            for b, branch_data in branch_map.items():
                key = tuple(
                    get_branch_value(b, branch_data, name)
                    for name, _, _ in workflow_params
                )
                if key not in branch_map_reversed:
                    branch_map_reversed[key] = []
                branch_map_reversed[key].append(b)
            # cache
            if h:
                cls._cls_branch_map_cache[h] = (branch_map, branch_map_reversed)

        # get parameters
        branch = params.get("branch", -1)
        branches = params.get("branches", ())

        # check if any or all workflow parameters are set, and if any of them is a sequence
        set_idxs = [i for i, (_, _, value) in enumerate(workflow_params) if value != no_value]
        any_set = len(set_idxs) > 0
        all_set = len(set_idxs) == len(workflow_params)
        any_seq = any(isinstance(value, (tuple, list, set)) for _, _, value in workflow_params)

        # when all are set and none of them is a sequence, the workflow parameters can refer to
        # no branch (-> exception), one branch (-> assign it), or multiple branches (-> workflow)
        _branches = []
        if all_set and not any_seq:
            values = tuple(value for _, _, value in workflow_params)
            _branches = branch_map_reversed.get(values, [])
            if len(_branches) == 0:
                raise ValueError(
                    "workflow parameters {} do not match any branch in {}".format(
                        wparams_repr(), cls.__name__,
                    ),
                )

        if all_set and not any_seq and _branches and len(_branches) == 1:
            # when all are set and do not refer to any sequence,
            # lookup the branch value and verify that workflow parameter values match
            _branch = _branches[0]
            if branch != -1 and branch != _branch:
                raise ValueError(
                    "workflow parameters {} in {} refer to branch {}, but branch {} "
                    "requested".format(wparams_repr(), cls.__name__, _branch, branch),
                )

            # always overwrite
            params["branch"] = branch = _branch

        elif any_set:
            # at least one parameter is not set or is a sequence, resulting in a workflow,
            # and in both cases we can filter the branch map to determine matching branches
            # branch should not be set
            if branch != -1:
                raise ValueError(
                    "workflow parameters {} will lead to {} being a workflow, but branch "
                    "{} requested".format(wparams_repr(), cls.__name__, branch),
                )

            if not _branches:
                # create a version of the reversed branch map where workflow parameters that are
                # not given are removed and corresponding branch values are merged
                branch_map_reversed_collapsed = defaultdict(list)
                for values, b in branch_map_reversed.items():
                    collapsed_values = tuple(values[i] for i in set_idxs)
                    branch_map_reversed_collapsed[collapsed_values].extend(b)

                # lookup all branches matched by parameters
                _branches = []
                names = [name for name, _, _ in workflow_params]
                sequences = (make_list(value) for _, _, value in workflow_params)
                for values in itertools.product(*sequences):
                    collapsed_values = tuple(values[i] for i in set_idxs)
                    if collapsed_values not in branch_map_reversed_collapsed:
                        param_repr = cjoin(map("{0[0]}={0[1]}".format, zip(names, values)))
                        raise Exception(
                            "workflow parameter combination {} not found in branch map of "
                            "{}".format(param_repr, cls.__name__),
                        )
                    _branches.extend(branch_map_reversed_collapsed[collapsed_values])

            # check if _branches match branches when set
            if branches:
                branches = range_expand(list(branches), include_end=True, min_value=0,
                    max_value=max(branch_map))
                if set(branches) != set(_branches):
                    raise ValueError(
                        "workflow parameters {} expanded in {} to branches ({}) do not match "
                        "passed branches ({})".format(
                            wparams_repr(), cls.__name__, cjoin(_branches), cjoin(branches),
                        ),
                    )

            # always overwrite
            params["branches"] = tuple(range_join(_branches))

        elif branch != -1:
            # set all workflow parameters according to the data in the branch map at "branch"
            if branch not in branch_map:
                raise KeyError(
                    "branch map of task class {} does not contain branch {}".format(
                        cls.__name__, branch,
                    ),
                )

            branch_data = branch_map[branch]
            for name, _, _ in workflow_params:
                params[name] = get_branch_value(branch, branch_data, name)

        return params

    @classmethod
    def find_workflow_cls(cls, name=None, fallback_to_first=False):
        first_cls = None

        for workflow_cls in inspect.getmro(cls):
            if not issubclass(workflow_cls, BaseWorkflow):
                continue
            if not workflow_cls._defined_workflow_proxy:
                continue
            if name in (workflow_cls.workflow_proxy_cls.workflow_type, None, NO_STR):
                return workflow_cls
            if first_cls is None:
                first_cls = workflow_cls

        if fallback_to_first and first_cls is not None:
            return first_cls

        msg = " for type '{}'".format(name) if name else ""
        raise ValueError("cannot determine workflow class{} in task class {}".format(msg, cls))

    @classmethod
    def _sanitize_branch_map(cls, branch_map, force_contiguous_branches):
        if isinstance(branch_map, (list, tuple)):
            branch_map = dict(enumerate(branch_map))
        elif isinstance(branch_map, six.integer_types):
            branch_map = dict(enumerate(range(branch_map)))
        elif force_contiguous_branches:
            n = len(branch_map)
            if set(branch_map.keys()) != set(range(n)):
                raise ValueError("branch map keys must constitute contiguous range "
                    "[0, {})".format(n))
        else:
            for branch in branch_map:
                if not isinstance(branch, six.integer_types) or branch < 0:
                    raise ValueError("branch map keys must be non-negative integers, got "
                        "'{}' ({})".format(branch, type(branch).__name__))

        return branch_map

    @classmethod
    def req_different_branching(cls, inst, **kwargs):
        """
        Variation of :py:meth:`Task.req` that should be used when defining requirements between
        workflows that implement a different branch granularity (e.g. task B with 10 branches
        requires task A with 2 branches). The only difference to the base method is that workflow
        specific parameters such as *branches* or *tolerance* are automatically skipped when not
        added explicitly in *kwargs*.
        """
        _exclude = set(make_list(kwargs.get("_exclude", [])))
        _exclude |= cls.exclude_params_branch
        kwargs["_exclude"] = _exclude

        return cls.req(inst, **kwargs)

    def __init__(self, *args, **kwargs):
        super(BaseWorkflow, self).__init__(*args, **kwargs)

        # store a list of workflow parameter names
        self._workflow_param_names = [
            name
            for name, param in self.get_params()
            if isinstance(param, WorkflowParameter)
        ]

        # workflow and branch specific attributes
        if self.is_workflow():
            # caches
            self._branch_map = None
            self._branch_tasks = None
            self._cache_branch_map = self.__class__.cache_branch_map_default
            self._cached_workflow_requirements = no_value

            # store whether workflow objects have been setup, which is done lazily,
            # and predefine all attributes that are set by it
            self._workflow_initialized = False
            self._workflow_cls = None
            self._workflow_proxy = None

            # initially set branches
            self._initial_branches = tuple(self.branches)

        else:
            # caches
            self._workflow_task = None

    @workflow_property(attr="_cache_branch_map")
    def cache_branch_map(self):
        return self._cache_branch_map

    @property
    def _cache_branches(self):
        # deprecation warning until v0.1
        logger.warning(
            "accessing {0}._cache_branches is deprecated, use {0}.cache_branch_map instead".format(
                self.__class__.__name__,
            ),
        )
        return self._cache_branch_map

    @_cache_branches.setter
    def _cache_branches(self, cache_branches):
        logger.warning(
            "setting {0}._cache_branches is deprecated, use {0}.cache_branch_map instead".format(
                self.__class__.__name__,
            ),
        )
        self._cache_branch_map = cache_branches

    def _initialize_workflow(self, force=False):
        if self.is_branch():
            return

        if self._workflow_initialized and not force:
            return

        self._workflow_cls = self.find_workflow_cls(self.effective_workflow)
        self._workflow_proxy = self._workflow_cls.workflow_proxy_cls(task=self)
        logger.debug(
            "created workflow proxy instance of type '{}'".format(self.effective_workflow),
        )

        self._workflow_initialized = True

    @property
    def workflow_cls(self):
        self._initialize_workflow()
        return self.as_workflow()._workflow_cls

    @property
    def workflow_proxy(self):
        self._initialize_workflow()
        return self.as_workflow()._workflow_proxy

    def repr(self, *args, **kwargs):
        if self.create_branch_map_before_repr:
            self.get_branch_map()

        return super(BaseWorkflow, self).repr(*args, **kwargs)

    def cli_args(self, exclude=None, replace=None):
        exclude = set() if exclude is None else set(make_list(exclude))

        # exclude certain branch/workflow parameters
        exclude |= self.exclude_params_branch if self.is_branch() else self.exclude_params_workflow

        # always exclude workflow parameters
        exclude |= set(self._workflow_param_names)

        return super(BaseWorkflow, self).cli_args(exclude=exclude, replace=replace)

    def _repr_params(self, *args, **kwargs):
        params = super(BaseWorkflow, self)._repr_params(*args, **kwargs)

        if self.is_workflow():
            # when this is a workflow, add the requested or effective workflow type,
            # depending on whether the requested one is to be passed through
            workflow = (
                self.workflow
                if self.passthrough_requested_workflow
                else self.effective_workflow
            )
            params.setdefault("workflow", workflow)
            # skip branches when empty
            if not params.get("branches"):
                params.pop("branches", None)
        else:
            # when this is a branch, remove workflow parameters
            for param in self.exclude_params_branch:
                params.pop(param, None)

        return params

    def req_branch(self, branch, **kwargs):
        if branch == -1:
            raise ValueError(
                "branch must not be -1 when creating a new branch task via req_branch(), "
                "but got {}".format(branch),
            )

        # default kwargs
        kwargs.setdefault("_skip_task_excludes", True)
        kwargs["_exclude"] = make_set(kwargs.get("_exclude", ())) | set(self._workflow_param_names)
        if self.is_workflow():
            kwargs["_exclude"] |= set(self.exclude_params_branch)

        # create the task
        task = self.req(self, branch=branch, **kwargs)

        # set the _workflow_task attribute if known
        if task._workflow_task is None:
            task._workflow_task = self if self.is_workflow() else self._workflow_task

        return task

    def req_workflow(self, **kwargs):
        # default kwargs
        kwargs.setdefault("_skip_task_excludes", True)
        kwargs["_exclude"] = make_set(kwargs.get("_exclude", ())) | set(self._workflow_param_names)
        if self.is_branch():
            kwargs["_exclude"] |= set(self.exclude_params_workflow)

        return self.req(self, branch=-1, **kwargs)

    def is_branch(self):
        """
        Returns whether or not this task refers to a *branch*.
        """
        return self.branch != -1

    def is_workflow(self):
        """
        Returns whether or not this task refers to the *workflow*.
        """
        return not self.is_branch()

    def as_branch(self, branch=None):
        """
        When this task refers to the workflow, a re-instantiated task with identical parameters and
        a certain *branch* value, defaulting to 0, is returned. When this task is already a branch
        task, the task itself is returned when *branch* is *None* or matches this task's branch
        value. Otherwise, a new branch task with that value and identical parameters is created and
        returned.
        """
        if branch == -1:
            raise ValueError("branch must not be -1 when selecting a branch task")

        if self.is_branch() and branch in (None, self.branch):
            return self

        return self.req_branch(branch or 0)

    def as_workflow(self, **kwargs):
        """
        When this task refers to a branch task, a re-instantiated task with ``branch=-1`` and
        identical parameters is returned. Otherwise, the workflow itself is returned. All *kwargs*
        are passed to :py:meth:`req_workflow`.
        """
        if self.is_workflow():
            return self

        if self._workflow_task is None:
            self._workflow_task = self.req_workflow(**kwargs)

        return self._workflow_task

    @abstractmethod
    def create_branch_map(self):
        """
        Abstract method that must be overwritten by inheriting tasks to define the branch map.
        """
        return

    def _reset_branch_boundaries(self, full_branch_map):
        if self.is_branch():
            raise Exception("calls to _reset_branch_boundaries are forbidden for branch tasks")

        # rejoin branch ranges when given
        if self.branches:
            # get minimum and maximum branches
            branches = set(full_branch_map.keys())
            min_branch = min(branches)
            max_branch = max(branches) + 1

            # get expanded branch values
            branches = range_expand(
                list(self.branches),
                min_value=min_branch,
                max_value=max_branch,
            )

            # assign back to branches attribute, use an empty tuple in case all branches are used
            use_all = (
                len(branches) == len(full_branch_map) and
                set(branches) == set(full_branch_map)
            )
            self.branches = () if use_all else tuple(range_join(branches))

    def _reduce_branch_map(self, branch_map):
        if self.is_branch():
            raise Exception("calls to _reduce_branch_map are forbidden for branch tasks")

        # create a set of branches to remove
        remove_branches = set()

        # apply branch ranges
        if self.branches:
            branches = set(branch_map.keys())
            min_branch = min(branches)
            max_branch = max(branches) + 1

            requested = range_expand(
                list(self.branches),
                min_value=min_branch,
                max_value=max_branch,
            )
            remove_branches |= branches - set(requested)

        # remove from branch map
        for b in remove_branches:
            del branch_map[b]

    def get_branch_map(self, reset_boundaries=True, reduce_branches=True):
        """
        Creates and returns the branch map defined in :py:meth:`create_branch_map`. If
        *reset_boundaries* is *True*, the branch numbers and ranges defined in :py:attr:`branches`
        are rearranged to not exceed the actual branch map length. If *reduce_branches* is *True*,
        the branch map is additionally filtered accordingly. The branch map is cached internally.
        """
        if self.is_branch():
            return self.as_workflow().get_branch_map(
                reset_boundaries=reset_boundaries,
                reduce_branches=reduce_branches,
            )

        branch_map = self._branch_map
        if branch_map is None:
            # create a new branch map
            args = ()
            if is_classmethod(self.create_branch_map, self.__class__):
                params = OrderedDict([
                    (param_name, getattr(self, param_name))
                    for param_name, _ in self.get_params()
                ])
                args = (params,)
            branch_map = self.create_branch_map(*args)

            # some type and sanity checks
            branch_map = self._sanitize_branch_map(branch_map, self.force_contiguous_branches)

            # post-process
            if reset_boundaries:
                self._reset_branch_boundaries(branch_map)
            if reduce_branches:
                self._reduce_branch_map(branch_map)

            # cache it
            if self.cache_branch_map:
                self._branch_map = branch_map

        return branch_map

    @property
    def branch_map(self):
        return self.get_branch_map()

    @property
    def branch_data(self):
        if self.is_workflow():
            raise Exception("calls to branch_data are forbidden for workflow tasks")

        branch_map = self.get_branch_map()
        if self.branch not in branch_map:
            raise ValueError("invalid branch '{}', not found in branch map".format(self.branch))

        return branch_map[self.branch]

    def get_branch_tasks(self):
        """
        Returns a dictionary that maps branch numbers to instantiated branch tasks. As this might be
        computationally intensive, the return value is cached.
        """
        if self.is_branch():
            return self.as_workflow().get_branch_tasks()

        if self._branch_tasks is None:
            # get all branch tasks according to the map
            branch_tasks = OrderedDict()
            for b in self.get_branch_map():
                branch_tasks[b] = self.as_branch(branch=b)

            # return the task when we are not going to cache it
            if not self.cache_branch_map:
                return branch_tasks

            # cache it
            self._branch_tasks = branch_tasks

        return self._branch_tasks

    def get_branch_chunks(self, chunk_size):
        """
        Returns a list of chunks of branch numbers defined in this workflow with a certain
        *chunk_size*. Example:

        .. code-block:: python

            wf = SomeWorkflowTask()  # has 8 branches
            print(wf.get_branch_chunks(3))
            # -> [[0, 1, 2], [3, 4, 5], [6, 7]]

            wf2 = SomeWorkflowTask(branches=[(0, 5)])  # has 5 branches
            print(wf2.get_branch_chunks(3))
            # -> [[0, 1, 2], [3, 4]]
        """
        if self.is_branch():
            return self.as_workflow().get_branch_chunks(chunk_size)

        # get the branch map and create chunks of its branch values
        branch_chunks = iter_chunks(self.get_branch_map().keys(), chunk_size)

        return list(branch_chunks)

    def get_all_branch_chunks(self, chunk_size, **kwargs):
        """
        Returns a list of chunks of all branch numbers of this workflow (i.e. without
        *branches* parameters applied) with a certain *chunk_size*. Internally, a new instance of
        this workflow is created using :py:meth:`BaseTask.req`, forwarding all *kwargs*, with
        *_exclude* parameters extended by ``{"branches"}`` in order to use all possible branch
        values. Example:

        .. code-block:: python

            wf = SomeWorkflowTask()  # has 8 branches
            print(wf.get_all_branch_chunks(3))
            # -> [[0, 1, 2], [3, 4, 5], [6, 7]]

            wf2 = SomeWorkflowTask(branches=[(0, 5)])  # has 5 branches
            print(wf2.get_all_branch_chunks(3))
            # -> [[0, 1, 2], [3, 4, 5], [6, 7]]
        """
        if self.is_branch():
            return self.as_workflow().get_all_branch_chunks(chunk_size, **kwargs)

        # create a new workflow instance
        kwargs["_exclude"] = set(kwargs.get("_exclude", set())) | {"branches"}
        kwargs["_skip_task_excludes"] = True
        wf = self.req_workflow(**kwargs)

        # return its branch chunks
        return wf.get_branch_chunks(chunk_size)

    def get_branches_repr(self, max_ranges=10):
        """
        Creates a string representation of the selected branches that can be used as a readable
        description or postfix in output paths. When the branches of this workflow are configured
        via the *branches* parameter, and there are more than *max_ranges* identified ranges, the
        string will contain a unique hash describing those ranges.
        """
        branch_map = self.get_branch_map()
        if not branch_map:
            return ""

        if not self.branches:
            return "{}To{}".format(min(branch_map.keys()), max(branch_map.keys()) + 1)

        ranges = range_join(list(branch_map.keys()))
        if len(ranges) > max_ranges:
            return "{}_ranges_{}".format(len(ranges), create_hash(ranges))

        return "_".join(
            str(r[0]) if len(r) == 1 else "{}To{}".format(r[0], r[1] + 1)
            for r in ranges
        )

    def workflow_complete(self):
        """
        Hook to define the completeness status of the workflow.
        """
        return NotImplemented

    def workflow_requires(self):
        """
        Hook to add workflow requirements. This method is expected to return a dictionary. When
        this method is called from a branch task, an exception is raised.
        """
        if self.is_branch():
            return self.as_workflow().workflow_requires()

        return DotDict()

    def workflow_input(self):
        """
        Returns the output targets of all workflow requirements, comparable to the normal
        ``input()`` method of plain tasks.
        """
        if self.is_branch():
            return self.as_workflow().workflow_input()

        # get potentially cached workflow requirements
        if self.cache_workflow_requirements:
            if self._cached_workflow_requirements is no_value:
                self._cached_workflow_requirements = self.workflow_proxy.requires()
            reqs = self._cached_workflow_requirements
        else:
            reqs = self.workflow_proxy.requires()

        return luigi.task.getpaths(reqs)

    def requires_from_branch(self):
        """
        Returns the requirements defined in the standard ``requires()`` method, but called in the
        context of the workflow. This method is only recommended in case all required tasks that
        would normally take a branch number, are intended to be instantiated with ``branch=-1``.
        When this method is called from a branch task, an exception is raised.
        """
        if self.is_branch():
            raise Exception("calls to requires_from_branch are forbidden for branch tasks")

        return self.__class__.requires(self)

    def _handle_scheduler_messages(self):
        if self.scheduler_messages:
            while not self.scheduler_messages.empty():
                msg = self.scheduler_messages.get()
                self.handle_scheduler_message(msg)

    def handle_scheduler_message(self, msg, _attr_value=None):
        """ handle_scheduler_message(msg)
        Hook that is called when a scheduler message *msg* is received. Returns *True* when the
        messages was handled, and *False* otherwise.

        Handled messages:

            - ``tolerance = <int/float>``
            - ``acceptance = <int/float>``
        """
        attr, value = _attr_value or (None, None)

        # handle "tolerance"
        if attr is None:
            m = re.match(r"^\s*(tolerance)\s*(\=|\:)\s*(.*)\s*$", str(msg))
            if m:
                attr = "tolerance"
                try:
                    self.tolerance = float(m.group(3))
                    value = self.tolerance
                except ValueError as e:
                    value = e

        # handle "acceptance"
        if attr is None:
            m = re.match(r"^\s*(acceptance)\s*(\=|\:)\s*(.*)\s*$", str(msg))
            if m:
                attr = "acceptance"
                try:
                    self.acceptance = float(m.group(3))
                    value = self.acceptance
                except ValueError as e:
                    value = e

        # respond
        if attr:
            if isinstance(value, Exception):
                msg.respond("cannot set {}: {}".format(attr, value))
                logger.warning("cannot set {} of task {}: {}".format(attr, self.live_task_id, value))
            else:
                msg.respond("{} set to {}".format(attr, value))
                logger.info("{} of task {} set to {}".format(attr, self.live_task_id, value))
            return True

        msg.respond("task cannot handle scheduler message: {}".format(msg))
        return False



# coding: utf-8

"""
gLite remote workflow implementation. See
https://wiki.italiangrid.it/twiki/bin/view/CREAM/UserGuide.
"""

__all__ = ["GLiteWorkflow"]


import os
import sys
import contextlib
from abc import abstractmethod
from collections import OrderedDict

import six

import law
from law.config import Config
from law.workflow.remote import BaseRemoteWorkflow, BaseRemoteWorkflowProxy
from law.job.base import JobArguments, JobInputFile, DeprecatedInputFiles
from law.task.proxy import ProxyCommand
from law.target.file import get_path
from law.parameter import CSVParameter
from law.util import no_value, law_src_path, merge_dicts, DotDict
from law.logger import get_logger

from law.contrib.glite.job import GLiteJobManager, GLiteJobFileFactory


logger = get_logger(__name__)


class GLiteWorkflowProxy(BaseRemoteWorkflowProxy):

    workflow_type = "glite"

    def __init__(self, *args, **kwargs):
        super(GLiteWorkflowProxy, self).__init__(*args, **kwargs)

        # check if there is at least one ce
        if not self.task.glite_ce:
            raise Exception("please set at least one glite computing element (--glite-ce)")

        self.delegation_ids = None

    def create_job_manager(self, **kwargs):
        return self.task.glite_create_job_manager(**kwargs)

    def setup_job_mananger(self):
        kwargs = {}

        # delegate the voms proxy to all endpoints
        if callable(self.task.glite_delegate_proxy):
            delegation_ids = []
            for ce in self.task.glite_ce:
                endpoint = law.wlcg.get_ce_endpoint(ce)
                delegation_ids.append(self.task.glite_delegate_proxy(endpoint))
            kwargs["delegation_id"] = delegation_ids

        return kwargs

    def create_job_file_factory(self, **kwargs):
        return self.task.glite_create_job_file_factory(**kwargs)

    def create_job_file(self, job_num, branches):
        task = self.task

        # the file postfix is pythonic range made from branches, e.g. [0, 1, 2, 4] -> "_0To5"
        postfix = "_{}To{}".format(branches[0], branches[-1] + 1)

        # create the config
        c = self.job_file_factory.get_config()
        c.input_files = DeprecatedInputFiles()
        c.output_files = []
        c.render_variables = {}
        c.custom_content = []

        # get the actual wrapper file that will be executed by the remote job
        wrapper_file = task.glite_wrapper_file()
        law_job_file = task.glite_job_file()
        if wrapper_file and get_path(wrapper_file) != get_path(law_job_file):
            c.input_files["executable_file"] = wrapper_file
            c.executable = wrapper_file
        else:
            c.executable = law_job_file
        c.input_files["job_file"] = law_job_file

        # collect task parameters
        exclude_args = (
            task.exclude_params_branch |
            task.exclude_params_workflow |
            task.exclude_params_remote_workflow |
            task.exclude_params_glite_workflow |
            {"workflow", "effective_workflow"}
        )
        proxy_cmd = ProxyCommand(
            task.as_branch(branches[0]),
            exclude_task_args=exclude_args,
            exclude_global_args=["workers", "local-scheduler", task.task_family + "-*"],
        )
        if task.glite_use_local_scheduler():
            proxy_cmd.add_arg("--local-scheduler", "True", overwrite=True)
        for key, value in OrderedDict(task.glite_cmdline_args()).items():
            proxy_cmd.add_arg(key, value, overwrite=True)

        # job script arguments
        job_args = JobArguments(
            task_cls=task.__class__,
            task_params=proxy_cmd.build(skip_run=True),
            branches=branches,
            workers=task.job_workers,
            auto_retry=False,
            dashboard_data=self.dashboard.remote_hook_data(
                job_num, self.job_data.attempts.get(job_num, 0)),
        )
        c.arguments = job_args.join()

        # add the bootstrap file
        bootstrap_file = task.glite_bootstrap_file()
        if bootstrap_file:
            c.input_files["bootstrap_file"] = bootstrap_file

        # add the stageout file
        stageout_file = task.glite_stageout_file()
        if stageout_file:
            c.input_files["stageout_file"] = stageout_file

        # does the dashboard have a hook file?
        dashboard_file = self.dashboard.remote_hook_file()
        if dashboard_file:
            c.input_files["dashboard_file"] = dashboard_file

        # initialize logs with empty values and defer to defaults later
        c.stdout = no_value
        c.stderr = no_value
        if task.transfer_logs:
            log_file = "stdall.txt"
            c.stdout = log_file
            c.stderr = log_file
            c.custom_log_file = log_file

        # meta infos
        c.output_uri = task.glite_output_uri()

        # task hook
        c = task.glite_job_config(c, job_num, branches)

        # build the job file and get the sanitized config
        job_file, c = self.job_file_factory(postfix=postfix, **c.__dict__)

        # logging defaults
        c.stdout = c.stdout or None
        c.stderr = c.stderr or None
        c.custom_log_file = c.custom_log_file or None

        # determine the custom log file uri if set
        abs_log_file = None
        if c.custom_log_file:
            abs_log_file = os.path.join(str(c.output_uri), c.custom_log_file)

        # return job and log files
        return {"job": job_file, "config": c, "log": abs_log_file}

    def destination_info(self):
        info = super(GLiteWorkflowProxy, self).destination_info()

        info["ce"] = "ce: {}".format(",".join(self.task.glite_ce))

        info = self.task.glite_destination_info(info)

        return info


class GLiteWorkflow(BaseRemoteWorkflow):

    workflow_proxy_cls = GLiteWorkflowProxy

    glite_workflow_run_decorators = None
    glite_job_manager_defaults = None
    glite_job_file_factory_defaults = None

    glite_ce = CSVParameter(
        default=(),
        significant=False,
        description="target glite computing element(s); default: empty",
    )

    glite_job_kwargs = []
    glite_job_kwargs_submit = ["glite_ce"]
    glite_job_kwargs_cancel = None
    glite_job_kwargs_cleanup = None
    glite_job_kwargs_query = None

    exclude_params_branch = {"glite_ce"}

    exclude_params_glite_workflow = set()

    exclude_index = True

    @contextlib.contextmanager
    def glite_workflow_run_context(self):
        """
        Hook to provide a context manager in which the workflow run implementation is placed. This
        can be helpful in situations where resurces should be acquired before and released after
        running a workflow.
        """
        yield

    @abstractmethod
    def glite_output_directory(self):
        return None

    @abstractmethod
    def glite_bootstrap_file(self):
        return None

    def glite_wrapper_file(self):
        return None

    def glite_job_file(self):
        return JobInputFile(law_src_path("job", "law_job.sh"))

    def glite_stageout_file(self):
        return None

    def glite_workflow_requires(self):
        return DotDict()

    def glite_output_postfix(self):
        return ""

    def glite_output_uri(self):
        return self.glite_output_directory().uri()

    def glite_job_resources(self, job_num, branches):
        """
        Hook to define resources for a specific job with number *job_num*, processing *branches*.
        This method should return a dictionary.
        """
        return {}

    def glite_delegate_proxy(self, endpoint):
        return law.wlcg.delegate_vomsproxy_glite(endpoint, stdout=sys.stdout, stderr=sys.stderr,
            cache=True)

    def glite_job_manager_cls(self):
        return GLiteJobManager

    def glite_create_job_manager(self, **kwargs):
        kwargs = merge_dicts(self.glite_job_manager_defaults, kwargs)
        return self.glite_job_manager_cls()(**kwargs)

    def glite_job_file_factory_cls(self):
        return GLiteJobFileFactory

    def glite_create_job_file_factory(self, **kwargs):
        # get the file factory cls
        factory_cls = self.glite_job_file_factory_cls()

        # job file fectory config priority: kwargs > class defaults
        kwargs = merge_dicts({}, self.glite_job_file_factory_defaults, kwargs)

        # default mkdtemp value which might require task-level info
        if kwargs.get("mkdtemp") is None:
            cfg = Config.instance()
            mkdtemp = cfg.get_expanded(
                "job",
                cfg.find_option("job", "glite_job_file_dir_mkdtemp", "job_file_dir_mkdtemp"),
            )
            if isinstance(mkdtemp, six.string_types) and mkdtemp.lower() not in {"true", "false"}:
                kwargs["mkdtemp"] = factory_cls._expand_template_path(
                    mkdtemp,
                    variables={"task_id": self.live_task_id, "task_family": self.task_family},
                )

        return factory_cls(**kwargs)

    def glite_job_config(self, config, job_num, branches):
        return config

    def glite_dump_intermediate_job_data(self):
        """
        Whether to dump intermediate job data to the job submission file while jobs are being
        submitted.
        """
        return True

    def glite_post_submit_delay(self):
        """
        Configurable delay in seconds to wait after submitting jobs and before starting the status
        polling.
        """
        return self.poll_interval * 60

    def glite_check_job_completeness(self):
        return False

    def glite_check_job_completeness_delay(self):
        return 0.0

    def glite_poll_callback(self, poll_data):
        """
        Configurable callback that is called after each job status query and before potential
        resubmission. It receives the variable polling attributes *poll_data* (:py:class:`PollData`)
        that can be changed within this method.

        If *False* is returned, the polling loop is gracefully terminated. Returning any other value
        does not have any effect.
        """
        return

    def glite_use_local_scheduler(self):
        return True

    def glite_cmdline_args(self):
        return {}

    def glite_destination_info(self, info):
        return info



# coding: utf-8

"""
Coffea target formatters.
"""

__all__ = ["CoffeaFormatter"]


from law.target.formatter import Formatter
from law.target.file import get_path
from law.logger import get_logger
from law.util import no_value


logger = get_logger(__name__)


class CoffeaFormatter(Formatter):

    name = "coffea"

    @classmethod
    def accepts(cls, path, mode):
        return get_path(path).endswith((".coffea", ".root", ".parquet"))

    @classmethod
    def load(cls, path, *args, **kwargs):
        path = get_path(path)

        if path.endswith(".root"):
            from coffea.nanoevents import NanoEventsFactory
            return NanoEventsFactory.from_root(path, *args, **kwargs)

        if path.endswith(".parquet"):
            from coffea.nanoevents import NanoEventsFactory
            return NanoEventsFactory.from_parquet(path, *args, **kwargs)

        # .coffea
        from coffea.util import load
        return load(path, *args, **kwargs)

    @classmethod
    def dump(cls, path, out, *args, **kwargs):
        from coffea.util import save

        perm = kwargs.pop("perm", no_value)

        save(out, get_path(path), *args, **kwargs)

        if perm != no_value:
            cls.chmod(path, perm)



# coding: utf-8

"""
Custom luigi parameters.
"""

__all__ = [
    "NO_STR", "NO_INT", "NO_FLOAT", "is_no_param", "get_param", "Parameter",
    "TaskInstanceParameter", "OptionalBoolParameter", "DurationParameter", "BytesParameter",
    "CSVParameter", "MultiCSVParameter", "RangeParameter", "MultiRangeParameter", "NotifyParameter",
    "NotifyMultiParameter", "NotifyMailParameter",
]

import functools
import csv

import luigi
import six

from law.notification import notify_mail
from law.util import (
    human_duration, parse_duration, time_units, time_unit_aliases, human_bytes, parse_bytes,
    byte_units, is_lazy_iterable, make_tuple, make_unique, brace_expand, range_expand, try_int,
    no_value,
)
from law.logger import get_logger


logger = get_logger(__name__)

# make luigi's BoolParameter parsing explicit globally, https://github.com/spotify/luigi/pull/2427
luigi.BoolParameter.parsing = getattr(luigi.BoolParameter, "EXPLICIT_PARSING", "explicit")

# also update existing BoolParameter instances in luigi's config classes to have explicit parsing
for cls in luigi.task.Config.__subclasses__():
    for attr in dir(cls):
        member = getattr(cls, attr)
        if isinstance(member, luigi.BoolParameter):
            member.parsing = luigi.BoolParameter.parsing

#: String value denoting an empty parameter.
NO_STR = "NO_STR"

#: Integer value denoting an empty parameter.
NO_INT = -1

#: Float value denoting an empty parameter.
NO_FLOAT = -1.0


def is_no_param(value):
    """
    Checks whether a parameter *value* denotes an empty parameter, i.e., if the value is either
    :py:attr:`NO_STR`, :py:attr:`NO_INT`, or :py:attr:`NO_FLOAT`.
    """
    return value in (NO_STR, NO_INT, NO_FLOAT, no_value)


def get_param(value, default=None):
    """
    Returns the passed *value* when it does not refer to an empty parameter value, checked with
    :py:func:`is_no_param`. Otherwise, *default* is returned, which defaults to *None*.
    """
    return default if is_no_param(value) else value


class Parameter(luigi.Parameter):
    """ __init__(*args, parse_empty=False, **kwargs)
    Custom base class of law-based parameters that adds additional features.

    As per luigi's default behavior, passing empty strings on the command line for a parameter leads
    to its default value and parsing is not triggered. When *parse_empty* is *True*, the parser is
    still called and should implement a custom behavior.

    All other *args* and *kwargs* are passed to :py:class:`luigi.Parameter`.

    .. py:attribute:: parse_empty

        type: bool

        Whether the parameter parsing should be triggered for empty command line arguments (usually
        empty strings, but not for *None*s).
    """

    def __init__(self, *args, **kwargs):
        self.parse_empty = kwargs.pop("parse_empty", False)

        super(Parameter, self).__init__(*args, **kwargs)


class TaskInstanceParameter(Parameter):
    """
    Parameter that can be used to pass the instance of a task. This class does not implement
    parameter value parsing.
    """

    def serialize(self, x):
        """"""
        if isinstance(x, Task):
            return getattr(x, "live_task_id", x.task_id)
        return str(x)


class OptionalBoolParameter(luigi.BoolParameter, Parameter):
    """
    Same as luigi's ``BoolParameter`` (and unlike luigi's ``OptionalBoolParameter``) but parses and
    serializes ``"None"`` strings transparently to *None* values and vice-versa.
    """

    def parse(self, inp):
        """"""
        if isinstance(inp, bool) or inp is None:
            return inp

        s = str(inp).lower()
        if s in ("true", "yes", "1"):
            return True
        if s in ("false", "no", "0"):
            return False
        if s == "none":
            return None

        raise ValueError("cannot interpret '{}' as boolean".format(inp))


class DurationParameter(Parameter):
    """ __init__(unit="s", *args, **kwargs)
    Parameter that interprets a string (or float) value as a duration, represented by a float
    number with a configurable unit. *unit* is forwarded as both the *unit* and *input_unit*
    argument of :py:func:`law.util.parse_duration` which is used for the conversion. For optimal
    precision, value serialization uses :py:func:`law.util.human_duration` with *colon_format*.
    Example:

    .. code-block:: python

        p = DurationParameter(unit="s")
        p.parse("5")                      # -> 5.0 (using the unit implicitly)
        p.parse("5s")                     # -> 5.0
        p.parse("5m")                     # -> 300.0
        p.parse("05:10")                  # -> 310.0
        p.parse("5 minutes, 15 seconds")  # -> 310.0
        p.serialize(310)                  # -> "05:15"

        p = DurationParameter(unit="m")
        p.parse("5")                      # -> 5.0 (using the unit implicitly)
        p.parse("5s")                     # -> 0.083
        p.parse("5m")                     # -> 5.0
        p.parse("05:10")                  # -> 5.167
        p.parse("5 minutes, 15 seconds")  # -> 5.25
        p.serialize(310)                  # -> "05:15:00"

    For more info, see :py:func:`law.util.parse_duration` and :py:func:`law.util.human_duration`.
    """

    def __init__(self, *args, **kwargs):
        # validate and set the unit
        self._unit = None
        self.unit = kwargs.pop("unit", "s")

        super(DurationParameter, self).__init__(*args, **kwargs)

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, unit):
        unit = time_unit_aliases.get(unit, unit)
        if unit not in time_units:
            raise ValueError("unknown unit '{}', valid values are {}".format(
                unit, time_units.keys()))

        self._unit = unit

    def parse(self, inp):
        """"""
        if inp in (None, "", NO_STR, no_value):
            inp = "0"

        return parse_duration(inp, input_unit=self.unit, unit=self.unit)

    def serialize(self, value):
        """"""
        if not value:
            value = 0

        value_seconds = parse_duration(value, input_unit=self.unit, unit="s")
        return human_duration(seconds=value_seconds, colon_format=True)


class BytesParameter(Parameter):
    """ __init__(unit="MB", *args, **kwargs)
    Parameter that interprets a string (or float) value as a number of bytes, represented by a float
    number with a configurable unit. *unit* is forwarded as both the *unit* and *input_unit*
    argument of :py:func:`law.util.parse_bytes` which is used for the conversion.
    Example:

    .. code-block:: python

        p = BytesParameter(unit="MB")
        p.parse("5")                      # -> 5.0 (using the unit implicitly)
        p.parse("5 MB")                   # -> 5.0
        p.parse(5)                        # -> 5.0
        p.parse("1 GB")                   # -> 1024.0
        p.serialize(310)                  # -> "310MB"

        p = BytesParameter(unit="GB")
        p.parse("5")                      # -> 5.0 (using the unit implicitly)
        p.parse("1024 MB")                # -> 1.0
        p.serialize("2048 MB")            # -> "2GB"

    For more info, see :py:func:`law.util.parse_bytes` and :py:func:`law.util.human_bytes`.
    """

    def __init__(self, *args, **kwargs):
        # validate and set the unit
        self._unit = None
        self.unit = kwargs.pop("unit", "MB")

        super(BytesParameter, self).__init__(*args, **kwargs)

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, unit):
        if unit not in byte_units:
            raise ValueError("unknown unit '{}', valid values are {}".format(unit, byte_units))

        self._unit = unit

    def parse(self, inp):
        """"""
        if inp in (None, "", NO_STR, no_value):
            inp = "0"

        return parse_bytes(inp, input_unit=self.unit, unit=self.unit)

    def serialize(self, value):
        """"""
        if not value:
            value = 0

        value_bytes = parse_bytes(value, input_unit=self.unit, unit="bytes")
        v, u = human_bytes(value_bytes, unit=self.unit)

        return "{}{}".format(try_int(v), u)


class CSVParameter(Parameter):
    r""" __init__(*args, cls=luigi.Parameter, inst=None, unique=False, sort=False, min_len=None, \
        max_len=None, choices=None, brace_expand=False, escape_sep=True, force_tuple=True, **kwargs)
    Parameter that parses a comma-separated value (CSV) and produces a tuple. *cls* (*inst*) can
    refer to an other parameter class (instance) that will be used to parse and serialize the
    particular items.

    When *unique* is *True*, both parsing and serialization methods make sure that values are
    unique. *sort* can be a boolean or a function for sorting parameter values.

    When *min_len* (*max_len*) is set to an integer, an error is raised in case the number of
    elements to serialize or parse (evaluated after potentially ensuring uniqueness) deceeds
    (exceeds) that value. Just like in luigi's *ChoiceParamater*, *choices* can be a sequence of
    accepted values.

    When *brace_expand* is *True*, brace expansion is applied, potentially extending the list of
    values. However, note that in this case commas that are not meant to act as a delimiter cannot
    be quoted in csv-style with double quotes, but they should rather be backslash-escaped instead.
    Unless *escape_sep* is *False*, escaped separators (comma) are not split when parsing
    strings and, likewise, separators contained in values to serialze are escaped.

    By default, single values are parsed such that they result in a tuple containing a single item.
    However, when *force_tuple* is *False*, single values that do not end with a comma are not
    wrapped by a tuple. Likewise, during serialization they are converted to a string as is, whereas
    tuple containing only a single item will end with a trailing comma.

    Example:

    .. code-block:: python

        p = CSVParameter(cls=luigi.IntParameter)
        p.parse("4,5,6,6")
        # => (4, 5, 6, 6)
        p.serialize((7, 8, 9))
        # => "7,8,9"

        # "," that should not be used as delimiter
        p = CSVParameter()
        p.parse("a,b,\"c,d\"")
        # -> ("a", "b", "c,d")
        # same as
        p.parse("a,b,c\,d")
        # -> ("a", "b", "c,d")

        # uniqueness check
        p = CSVParameter(cls=luigi.IntParameter, unique=True)
        p.parse("4,5,6,6")
        # => (4, 5, 6)

        # length check
        p = CSVParameter(cls=luigi.IntParameter, max_len=2)
        p.parse("4,5,6")
        # => ValueError

        # choices
        p = CSVParameter(cls=luigi.IntParameter, choices=(1, 2))
        p.parse("2,3")
        # => ValueError

        # brace expansion
        p = CSVParameter(cls=luigi.IntParameter, brace_expand=True)
        # (note that with brace_expand enabled, the quoting if "," only works with back slashes)
        p.parse("1{2,3,4}9")
        # => (129, 139, 149)

        # do not force tuples to wrap single values
        p = CSVParameter(cls=luigi.IntParameter, force_tuple=False)
        p.parse("1")
        # => 1
        # note: the result would be (1,) with force_tuple left at True (default)
        p.parse("1,")
        # => (1,)
        p.serialize(1)
        # => "1"
        p.serialize((1,))
        # => "1,"
        p.serialize((1, 2))
        # => "1,2"

    .. note::

        Due to the way `instance caching
        <https://luigi.readthedocs.io/en/stable/parameters.html#parameter-instance-caching>`__
        is implemented in luigi, parameters should always have hashable, immutable values.
        Therefore, this parameter produces a tuple and, in particular, not a list. To avoid
        undesired side effects, the *default* value given to the constructor is also converted to a
        tuple.

    .. py:attribute:: _inst

        type: :py:attr:`cls`

        Instance of the luigi parameter class *cls* or *inst* directory, that is used internally for
        parameter parsing and serialization.
    """

    def __init__(self, *args, **kwargs):
        self._cls = kwargs.pop("cls", luigi.Parameter)
        self._inst = kwargs.pop("inst", None)
        self._unique = kwargs.pop("unique", False)
        self._sort = kwargs.pop("sort", False)
        self._min_len = kwargs.pop("min_len", None)
        self._max_len = kwargs.pop("max_len", None)
        self._choices = kwargs.pop("choices", None)
        self._brace_expand = kwargs.pop("brace_expand", False)
        self._escape_sep = kwargs.pop("escape_sep", True)
        self._force_tuple = kwargs.pop("force_tuple", True)

        # ensure that the default value is a tuple
        if "default" in kwargs:
            kwargs["default"] = make_tuple(kwargs["default"])

        # instantiate cls when inst is not set, or set cls base on inst
        if self._inst is None:
            self._inst = self._cls()
        else:
            self._cls = self._inst.__class__

        super(CSVParameter, self).__init__(*args, **kwargs)

    def _check_unique(self, value):
        if not self._unique:
            return value

        return make_unique(value)

    def _check_sort(self, value):
        if not self._sort:
            return value

        key = self._sort if callable(self._sort) else None
        value = sorted(value, key=key)

        return tuple(value)

    def _check_len(self, value):
        str_repr = lambda: ",".join(str(v) for v in value)

        if self._min_len is not None and len(value) < self._min_len:
            raise ValueError("'{}' contains {} value(s), a minimum of {} is required".format(
                str_repr(), len(value), self._min_len))

        # check max_len
        if self._max_len is not None and len(value) > self._max_len:
            raise ValueError("'{}' contains {} value(s), a maximum of {} is required".format(
                str_repr(), len(value), self._max_len))

    def _check_choices(self, value):
        if not self._choices:
            return

        unknown = []
        for v in value:
            if v not in self._choices:
                unknown.append(v)

        if unknown:
            str_repr = lambda value: ",".join(str(v) for v in value)
            raise ValueError("invalid parameter value(s) '{}', valid choices are '{}'".format(
                str_repr(make_unique(unknown)), str_repr(self._choices)))

    def parse(self, inp):
        """"""
        return_single_value = False
        if inp in (None, "", NO_STR, no_value):
            value = tuple()
        elif isinstance(inp, (tuple, list)) or is_lazy_iterable(inp):
            value = make_tuple(inp)
        elif isinstance(inp, six.string_types):
            if self._brace_expand:
                elems = brace_expand(inp, split_csv=True, escape_csv_sep=self._escape_sep)
            else:
                # replace escaped separators
                if self._escape_sep:
                    escaped_sep = "__law_escaped_csv_sep__"
                    inp = inp.replace("\\,", escaped_sep)
                # proper csv split
                elems = list(csv.reader([inp]))[0]
                # add back escaped separators per element
                if self._escape_sep:
                    elems = [elem.replace(escaped_sep, ",") for elem in elems]
            # skip trailing empty strings
            if not elems[-1]:
                elems.pop()
            # parse
            value = tuple(map(self._inst.parse, elems))
            return_single_value = len(value) == 1 and not self._force_tuple
        else:
            value = (inp,)

        # apply uniqueness, sort, length and choices checks
        value = self._check_unique(value)
        value = self._check_sort(value)
        self._check_len(value)
        self._check_choices(value)

        return value[0] if return_single_value else value

    def serialize(self, value):
        """"""
        if value in (None, NO_STR, no_value):
            value = tuple()

        # ensure value is a tuple
        was_sequence = isinstance(value, (tuple, list))
        value = make_tuple(value)

        # apply uniqueness, sort, length and choices checks
        value = self._check_unique(value)
        value = self._check_sort(value)
        self._check_len(value)
        self._check_choices(value)

        # convert to string
        s = ",".join(str(self._inst.serialize(elem)) for elem in value)

        # add a trailing comma if necessary
        if len(value) == 1 and not self._force_tuple and was_sequence:
            s += ","

        return s


class MultiCSVParameter(CSVParameter):
    r""" __init__(*args, cls=luigi.Parameter, inst=None, unique=False, sort=False, min_len=None, \
        max_len=None, choices=None, brace_expand=False, escape_sep=True, force_tuple=True, **kwargs)
    Parameter that parses several comma-separated values (CSV), separated by colons, and produces a
    nested tuple. *cls* (*inst*) can refer to an other parameter class (instance) that will be used
    to parse and serialize the particular items.

    Except for the additional support for multiple CSV sequences, the parsing and serialization
    implementation is based on :py:class:`CSVParameter`, which also handles the features controlled
    by *unique*, *sort*, *max_len*, *min_len*, *choices*, *brace_expand*, *escape_sep* and
    *force_tuple* per sequence of values.

    However, note that in this case colon characters that are not meant to act as a delimiter cannot
    be quoted in csv-style with double quotes, but they should rather be backslash-escaped instead.
    Unless *escape_sep* is *False*, escaped separators (colon) are not split when parsing strings
    and, likewise, separators contained in values to serialze are escaped.

    Example:

    .. code-block:: python

        p = MultiCSVParameter(cls=luigi.IntParameter)
        p.parse("4,5:6,6")
        # => ((4, 5), (6, 6))
        p.serialize((7, 8, (9,)))
        # => "7,8:9"

        # ":" that should not be used as delimiter
        p = MultiCSVParameter()
        p.parse("a,b:\"c:d\"")
        # -> (("a", "b"), ("c:d",))
        # same as
        p.parse("a,b:c\:d")
        # -> (("a", "b"), ("c:d",))

        # uniqueness check
        p = MultiCSVParameter(cls=luigi.IntParameter, unique=True)
        p.parse("4,5:6,6")
        # => ((4, 5), (6,))

        # length check
        p = MultiCSVParameter(cls=luigi.IntParameter, max_len=2)
        p.parse("4,5:6,7,8")
        # => ValueError

        # choices
        p = MultiCSVParameter(cls=luigi.IntParameter, choices=(1, 2))
        p.parse("1,2:2,3")
        # => ValueError

        # brace expansion
        # (note that with brace_expand enabled, the quoting if ":" only works with back slashes)
        p = MultiCSVParameter(cls=luigi.IntParameter, brace_expand=True)
        p.parse("4,5:6,7,8{8,9}")
        # => ((4, 5), (6, 7, 88, 89))

    .. note::

        Due to the way `instance caching
        <https://luigi.readthedocs.io/en/stable/parameters.html#parameter-instance-caching>`__
        is implemented in luigi, parameters should always have hashable values. Therefore, this
        parameter produces a (nested) tuple and, in particular, not a list. To avoid undesired side
        effects, the *default* value given to the constructor is also converted to a tuple.

    .. py:attribute:: _inst

        type: :py:attr:`cls`

        Instance of the luigi parameter class *cls* or *inst* directly, that is used internally for
        parameter parsing and serialization.
    """

    # custom csv dialect for splitting by ":" for automatic quoting
    class _Dialect(csv.excel):
        delimiter = ":"

    _dialect = _Dialect()

    def parse(self, inp):
        """"""
        if not inp or inp == NO_STR:
            value = tuple()
        elif isinstance(inp, (tuple, list)) or is_lazy_iterable(inp):
            value = tuple(super(MultiCSVParameter, self).parse(v) for v in inp)
        elif isinstance(inp, six.string_types):
            # replace escaped separators
            if self._escape_sep:
                escaped_sep = "__law_escaped_multi_csv_sep__"
                inp = inp.replace("\\" + ":", escaped_sep)
            # split
            elems = list(csv.reader([inp], dialect=self._dialect))[0]
            # add back escaped separators per element
            if self._escape_sep:
                elems = [elem.replace(escaped_sep, ":") for elem in elems]
            value = tuple(super(MultiCSVParameter, self).parse(e) for e in elems)
        else:
            value = (super(MultiCSVParameter, self).parse(inp),)

        return value

    def serialize(self, value):
        """"""
        if not value:
            return ""

        return ":".join(super(MultiCSVParameter, self).serialize(v) for v in make_tuple(value))


class RangeParameter(Parameter):
    """ __init__(*args, require_start=True, require_end=True, single_value=False, **kwargs)
    Parameter that parses a range in the format ``start:stop`` and returns a tuple with two integer
    elements.

    When *require_start* or *require_stop* are *False*, the formats ``:stop`` and ``start:``,
    respectively, are accepted as well. In these cases, the tuple will contain an attribute
    :py:attr:`OPEN` do denote that either side is unconstrained.

    When *single_value* is *True*, single integer values are accepted and lead to a tuple with one
    element.

    .. code-block:: python

        p = RangeParameter()
        p.parse("4:8")
        # => (4, 8)
        p.serialize((5, 9))
        # => "5:9"
        p.parse("4:")
        # => ValueError
        p.parse("4")
        # => ValueError

        p = RangeParameter(require_start=False, require_stop=False)
        p.parse("4:5")
        # => (4, 5)
        p.parse("4:")
        # => (4, OPEN)
        p.parse(":5")
        # => (OPEN, 5)
        p.parse(":")
        # => (OPEN, OPEN)
        p.serialize((OPEN, 8))
        # => ":8"

        p = RangeParameter(single_value=True)
        p.parse("4")
        # => (4,)
        p.serialize((5,))
        # => "4"

    .. py:classattribute:: RANGE_SEP

        type: string

        Character used as a separator between range edges when parsing strings and serializing
        values. Defaults to ``":"``.

    .. py:classattribute:: OPEN

        type: None

        Value denoting open edges in parsed ranges.
    """

    RANGE_SEP = ":"
    OPEN = None

    @classmethod
    def expand(cls, range, **kwargs):
        """
        Expands *range* (as returned by :py:meth:`parse`) to a sorted list of unique integers.
        Additional *kwargs* are forwarded to :py:func:`law.util.range_expand`.

        .. code-block:: python

            RangeParameter.expand((4, 8))
            # -> [4, 5, 6, 7]
        """
        return sorted(set(range_expand(range, **kwargs)))

    def __init__(self, *args, **kwargs):
        self._require_start = kwargs.pop("require_start", True)
        self._require_end = kwargs.pop("require_end", True)
        self._single_value = kwargs.pop("single_value", False)

        # ensure that the default value is a tuple
        if "default" in kwargs:
            kwargs["default"] = make_tuple(kwargs["default"])

        super(RangeParameter, self).__init__(*args, **kwargs)

    def _check(self, value):
        if not isinstance(value, tuple):
            raise TypeError("invalid type of range {}, must be a tuple".format(value))

        elif len(value) == 1 and self._single_value:
            if not isinstance(value[0], six.integer_types):
                raise TypeError("invalid type of single value in range {}".format(value))

        elif len(value) == 2:
            start, end = value
            if start == self.OPEN:
                if self._require_start:
                    raise ValueError("range {} lacks start value which is required".format(value))
            elif not isinstance(start, six.integer_types):
                raise TypeError("invalid type of start value in range {}".format(value))
            if end == self.OPEN:
                if self._require_end:
                    raise ValueError("range {} lacks end value which is required".format(value))
            elif not isinstance(end, six.integer_types):
                raise TypeError("invalid type of end value in range {}".format(value))

        elif value:
            raise ValueError("cannot interpret {} with {} elements as {}".format(
                value, len(value), self.__class__.__name__))

    def parse(self, inp):
        """"""
        if inp in (None, "", NO_STR, no_value):
            value = tuple()
        elif isinstance(inp, (tuple, list)) or is_lazy_iterable(inp):
            value = make_tuple(inp)
        elif isinstance(inp, six.integer_types):
            value = (inp,)
        elif isinstance(inp, six.string_types):
            parts = inp.split(self.RANGE_SEP)
            # convert integers
            try:
                value = tuple((int(p) if p else self.OPEN) for p in parts)
            except ValueError:
                raise ValueError("range '{}' contains non-integer elements".format(inp))
        else:
            value = (inp,)

        self._check(value)

        return value

    def serialize(self, value):
        """"""
        if not value:
            value = tuple()

        self._check(value)

        value = [("" if v == self.OPEN else str(v)) for v in value]
        return self.RANGE_SEP.join(value)


class MultiRangeParameter(RangeParameter):
    """ __init__(*args, require_start=True, require_end=True, single_value=False, **kwargs)
    Parameter that parses several integer ranges (each in the format ``start-end``), separated by
    comma, and produces a nested tuple.

    Except for the additional support for multiple ranges, the parsing and serialization
    implementation is based on :py:class:`RangeParameter`, which also handles the control of open
    edges with *require_start* and *require_end*, and the acceptance of single integer values with
    *single_value*.

    Example:

    .. code-block:: python

        p = MultiRangeParameter()
        p.parse("4:8,12:14")
        # => ((4, 8), (12, 14))
        p.serialize(((5, 9), (13, 15)))
        # => ""5:9,13:15""

    .. py:classattribute:: MULTI_RANGE_SEP

        type: string

        Character used as a separator between ranges when parsing strings and serializing
        values. Defaults to ``","``.
    """

    MULTI_RANGE_SEP = ","

    @classmethod
    def expand(cls, ranges, **kwargs):
        """
        Expands *ranges* (as returned by :py:meth:`parse`) to a sorted list of unique integers.
        Additional *kwargs* are forwarded to :py:func:`law.util.range_expand`.

        .. code-block:: python

            MultiRangeParameter.expand(((4, 8), (12, 14)))
            # -> [4, 5, 6, 7, 8, 12, 13, 14]
        """
        return sorted(set.union(*map(set, map(functools.partial(range_expand, **kwargs), ranges))))

    def parse(self, inp):
        """"""
        if inp in (None, "", NO_STR, no_value):
            value = tuple()
        elif isinstance(inp, (tuple, list)) or is_lazy_iterable(inp):
            value = tuple(super(MultiRangeParameter, self).parse(v) for v in inp)
        elif isinstance(inp, six.string_types):
            elems = inp.split(self.MULTI_RANGE_SEP)
            value = tuple(super(MultiRangeParameter, self).parse(e) for e in elems)
        else:
            value = (super(MultiRangeParameter, self).parse(inp),)

        return value

    def serialize(self, value):
        """"""
        if not value:
            return ""

        # ensure that value is a nested tuple
        value = tuple(map(make_tuple, make_tuple(value)))

        return self.MULTI_RANGE_SEP.join(
            super(MultiRangeParameter, self).serialize(v) for v in value)


class NotifyParameter(luigi.BoolParameter, Parameter):
    """
    Base class for notification parameters. A notification parameter must provide a notification
    transport in :py:meth:`get_transport`, e.g.

    .. code-block:: python

        def get_transport(self):
            return {
                "func": notification_func,
                "raw": True,  # or False
            }

    When a task has a specific notification parameter set to *True* and its run method is decorated
    with the :py:func:`law.notify` function, *notification_func* is called with at least three
    arguments: *success*, *title* and *message*. *success* is a boolean which is *True* when the
    decorated function did not raise an exception. *title* is always a string. When *raw* is *False*
    (the default), *message* is also a string. Otherwise, it is an ordered dictionary containing key
    value pairs describing the message content. All options passed to
    :py:func:`law.decorator.notify` are forwarded to *notification_func* as optional arguments.
    """

    def get_transport(self):
        """
        Method to configure the transport that is toggled by this parameter. Should return a
        dictionary with ``"func"`` and ``"raw"`` (optional) fields.
        """
        return None


class NotifyMultiParameter(NotifyParameter):
    """ __init__(parameters=[], *args, **kwargs)
    Parameter that takes multiple other :py:class:`NotifyParameter`'s to join their notification
    functionality in a single parameter. Example:

    .. code-block:: python

        class MyTask(law.Task):

            notify = law.NotifyMultiParameter(parameters=[
                law.NotifyMailParameter(significant=False),
                ...  # further NotifyParameters
            ])
    """

    def __init__(self, *args, **kwargs):
        self.parameters = kwargs.pop("parameters", [])

        super(NotifyMultiParameter, self).__init__(*args, **kwargs)

    def get_transport(self):
        """"""
        return [param.get_transport() for param in self.parameters]


class NotifyMailParameter(NotifyParameter):
    """
    Notification parameter defining a basic email transport. Uses
    :py:meth:`law.notification.notify_mail` internally.
    """

    def __init__(self, *args, **kwargs):
        super(NotifyMailParameter, self).__init__(*args, **kwargs)

        if not self.description:
            self.description = "when true, and the task's run method is decorated with " \
                "law.decorator.notify, an email notification is sent once the task finishes"

    @classmethod
    def notify(success, *args, **kwargs):
        """"""
        return notify_mail(*args, **kwargs)

    def get_transport(self):
        """"""
        return {
            "func": self.notify,
            "raw": False,
            "colored": False,
        }


# trailing imports
from law.task.base import Task



# coding: utf-8

"""
Singularity sandbox implementation.
"""

__all__ = ["SingularitySandbox"]


import os
import subprocess

import luigi
import six

from law.config import Config
from law.sandbox.base import Sandbox
from law.target.local import LocalDirectoryTarget, LocalFileTarget
from law.cli.software import get_software_deps
from law.util import make_list, interruptable_popen, quote_cmd, flatten, law_src_path, makedirs


class SingularitySandbox(Sandbox):

    sandbox_type = "singularity"

    config_section_prefix = sandbox_type

    @property
    def image(self):
        return self.name

    @property
    def env_cache_key(self):
        return self.image

    def get_custom_config_section_postfix(self):
        return self.image

    def create_env(self):
        # strategy: unlike docker, singularity might not allow binding of paths that do not exist
        # in the container, so create a tmp directory on the host system and bind it as /tmp, let
        # python dump its full env into a file, and read the file again on the host system

        # helper to load the env
        def load_env(target):
            try:
                return tmp.load(formatter="pickle")
            except Exception as e:
                raise Exception(
                    "env deserialization of sandbox {} failed: {}".format(self, e),
                )

        # load the env when the cache file is configured and existing
        if self.env_cache_path:
            env_cache_target = LocalFileTarget(self.env_cache_path)
            if env_cache_target.exists():
                return load_env(env_cache_target)

        # create tmp dir and file
        tmp_dir = LocalDirectoryTarget(is_tmp=True)
        tmp_dir.touch()
        tmp = tmp_dir.child("env", type="f")
        tmp.touch()

        # determine whether volume binding is allowed
        allow_binds_cb = getattr(self.task, "singularity_allow_binds", None)
        if callable(allow_binds_cb):
            allow_binds = allow_binds_cb()
        else:
            cfg = Config.instance()
            allow_binds = cfg.get_expanded(self.get_config_section(), "allow_binds")

        # arguments to configure the environment
        args = ["-e"]
        if allow_binds:
            args.extend(["-B", "{}:/tmp".format(tmp_dir.path)])
            env_file = "/tmp/{}".format(tmp.basename)
        else:
            env_file = tmp.path

        # get the singularity exec command
        singularity_exec_cmd = self._singularity_exec_cmd() + args

        # pre-setup commands
        pre_setup_cmds = self._build_pre_setup_cmds()

        # post-setup commands
        post_env = self._get_env()
        post_setup_cmds = self._build_post_setup_cmds(post_env)

        # build the python command that dumps the environment
        py_cmd = "import os,pickle;" \
            + "pickle.dump(dict(os.environ),open('{}','wb'),protocol=2)".format(env_file)

        # build the full command
        cmd = quote_cmd(singularity_exec_cmd + [self.image, "bash", "-l", "-c",
            " && ".join(flatten(
                pre_setup_cmds,
                post_setup_cmds,
                quote_cmd(["python", "-c", py_cmd]),
            )),
        ])

        # run it
        code, out, _ = interruptable_popen(cmd, shell=True, executable="/bin/bash",
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if code != 0:
            raise Exception(
                "singularity sandbox env loading failed with exit code {}:\n{}".format(
                    code, out),
            )

        # copy to the cache path when configured
        if self.env_cache_path:
            tmp.copy_to_local(env_cache_target)

        # load the env
        env = load_env(tmp)

        return env

    def _singularity_exec_cmd(self):
        cmd = ["singularity", "exec"]

        # task-specific argiments
        if self.task:
            # add args configured on the task
            args_getter = getattr(self.task, "singularity_args", None)
            if callable(args_getter):
                cmd.extend(make_list(args_getter()))

        return cmd

    def cmd(self, proxy_cmd):
        # singularity exec command arguments
        # -e clears the environment
        args = ["-e"]

        # helper to build forwarded paths
        cfg = Config.instance()
        cfg_section = self.get_config_section()
        forward_dir = cfg.get_expanded(cfg_section, "forward_dir")
        python_dir = cfg.get_expanded(cfg_section, "python_dir")
        bin_dir = cfg.get_expanded(cfg_section, "bin_dir")
        stagein_dir_name = cfg.get_expanded(cfg_section, "stagein_dir_name")
        stageout_dir_name = cfg.get_expanded(cfg_section, "stageout_dir_name")

        def dst(*args):
            return os.path.join(forward_dir, *(str(arg) for arg in args))

        # helper for mounting a volume
        volume_srcs = []

        def mount(*vol):
            src = vol[0]

            # make sure, the same source directory is not mounted twice
            if src in volume_srcs:
                return
            volume_srcs.append(src)

            # ensure that source directories exist
            if not os.path.isfile(src):
                makedirs(src)

            # store the mount point
            args.extend(["-B", ":".join(vol)])

        # determine whether volume binding is allowed
        allow_binds_cb = getattr(self.task, "singularity_allow_binds", None)
        if callable(allow_binds_cb):
            allow_binds = allow_binds_cb()
        else:
            allow_binds = cfg.get_expanded(cfg_section, "allow_binds")

        # determine whether law software forwarding is allowed
        forward_law_cb = getattr(self.task, "singularity_forward_law", None)
        if callable(forward_law_cb):
            forward_law = forward_law_cb()
        else:
            forward_law = cfg.get_expanded_bool(cfg_section, "forward_law")

        # environment variables to set
        env = self._get_env()

        # prevent python from writing byte code files
        env["PYTHONDONTWRITEBYTECODE"] = "1"

        if forward_law:
            # adjust path variables
            if allow_binds:
                env["PATH"] = os.pathsep.join([dst("bin"), "$PATH"])
                env["PYTHONPATH"] = os.pathsep.join([dst(python_dir), "$PYTHONPATH"])
            else:
                env["PATH"] = "$PATH"
                env["PYTHONPATH"] = "$PYTHONPATH"

            # forward python directories of law and dependencies
            for mod in get_software_deps():
                path = os.path.dirname(mod.__file__)
                name, ext = os.path.splitext(os.path.basename(mod.__file__))
                if name == "__init__":
                    vsrc = path
                    vdst = dst(python_dir, os.path.basename(path))
                else:
                    vsrc = os.path.join(path, name + ".py")
                    vdst = dst(python_dir, name + ".py")
                if allow_binds:
                    mount(vsrc, vdst)
                else:
                    dep_path = os.path.dirname(vsrc)
                    if dep_path not in env["PYTHONPATH"].split(os.pathsep):
                        env["PYTHONPATH"] = os.pathsep.join([dep_path, env["PYTHONPATH"]])

            # forward the law cli dir to bin as it contains a law executable
            if allow_binds:
                env["PATH"] = os.pathsep.join([dst(python_dir, "law", "cli"), env["PATH"]])
            else:
                env["PATH"] = os.pathsep.join([law_src_path("cli"), env["PATH"]])

            # forward the law config file
            if cfg.config_file:
                if allow_binds:
                    mount(cfg.config_file, dst("law.cfg"))
                    env["LAW_CONFIG_FILE"] = dst("law.cfg")
                else:
                    env["LAW_CONFIG_FILE"] = cfg.config_file

            # forward the luigi config file
            for p in luigi.configuration.LuigiConfigParser._config_paths[::-1]:
                if os.path.exists(p):
                    if allow_binds:
                        mount(p, dst("luigi.cfg"))
                        env["LUIGI_CONFIG_PATH"] = dst("luigi.cfg")
                    else:
                        env["LUIGI_CONFIG_PATH"] = p
                    break

        # add staging directories
        if (self.stagein_info or self.stageout_info) and not allow_binds:
            raise Exception("cannot use stage-in or -out if binds are not allowed")

        if self.stagein_info:
            env["LAW_SANDBOX_STAGEIN_DIR"] = dst(stagein_dir_name)
            mount(self.stagein_info.stage_dir.path, dst(stagein_dir_name))
        if self.stageout_info:
            env["LAW_SANDBOX_STAGEOUT_DIR"] = dst(stageout_dir_name)
            mount(self.stageout_info.stage_dir.path, dst(stageout_dir_name))

        # forward volumes defined in the config and by the task
        vols = self._get_volumes()
        if vols and not allow_binds:
            raise Exception("cannot forward volumes to sandbox if binds are not allowed")

        for hdir, cdir in six.iteritems(vols):
            if not cdir:
                mount(hdir)
            else:
                cdir = self._expand_volume(cdir, bin_dir=dst(bin_dir), python_dir=dst(python_dir))
                mount(hdir, cdir)

        # handle local scheduling within the container
        if self.force_local_scheduler():
            proxy_cmd.add_arg("--local-scheduler", "True", overwrite=True)

        # get the singularity exec command, add arguments from above
        singularity_exec_cmd = self._singularity_exec_cmd() + args

        # pre-setup commands
        pre_setup_cmds = self._build_pre_setup_cmds()

        # post-setup commands with the full env
        post_setup_cmds = self._build_post_setup_cmds(env)

        # build the final command
        cmd = quote_cmd(singularity_exec_cmd + [self.image, "bash", "-l", "-c",
            " && ".join(flatten(
                pre_setup_cmds,
                post_setup_cmds,
                proxy_cmd.build(),
            )),
        ])

        return cmd



# coding: utf-8

"""
Function returning the config defaults of the wlcg package.
"""

from copy import deepcopy


def config_defaults(default_config):
    defaults = {
        "target": {
            "default_wlcg_fs": "wlcg_fs",
        },
        "wlcg_fs": {
            # defined by FileSystem
            "has_permissions": False,
            "default_file_perm": None,
            "default_dir_perm": None,
            "create_file_dir": False,  # requires gfal_transfer_create_parent to be True
            # defined by RemoteFileInterface
            "base": None,
            "base_stat": None,
            "base_exists": None,
            "base_chmod": None,
            "base_unlink": None,
            "base_rmdir": None,
            "base_mkdir": None,
            "base_listdir": None,
            "base_filecopy": None,
            "retries": 1,
            "retry_delay": "5s",
            "random_base": True,
            # defined by RemoteFileSystem
            "validate_copy": False,
            "use_cache": False,
            # define by RemoteCache
            "cache_root": None,
            "cache_cleanup": None,
            "cache_max_size": "0MB",
            "cache_mtime_patience": 1.0,
            "cache_file_perm": 0o0660,
            "cache_dir_perm": 0o0770,
            "cache_wait_delay": "5s",
            "cache_max_waits": 120,
            "cache_global_lock": False,
            # defined by GFALFileInterface
            "gfal_atomic_contexts": False,
            "gfal_transfer_timeout": 3600,
            "gfal_transfer_checksum_check": False,
            "gfal_transfer_nbstreams": 1,
            "gfal_transfer_overwrite": True,
            "gfal_transfer_create_parent": True,
            "gfal_transfer_strict_copy": False,
            # defined by WLCGFileSystem
            # no dedicated configs
        },
    }

    defaults["_wlcg_fs_defaults"] = deepcopy(defaults["wlcg_fs"])

    return defaults



# coding: utf-8

"""
Functions that are invoked by interactive task methods.
"""

__all__ = [
    "print_task_deps", "print_task_status", "print_task_output", "remove_task_output",
    "fetch_task_output",
]


import os
import re

import six

from law.config import Config
from law.target.base import Target
from law.target.file import FileSystemTarget
from law.target.collection import TargetCollection, FileCollection
from law.util import (
    colored, uncolored, uncolor_cre, flatten, flag_to_bool, query_choice, human_bytes,
    is_lazy_iterable, make_list, merge_dicts, makedirs, get_terminal_width,
)
from law.logger import get_logger


logger = get_logger(__name__)


# formatting characters
fmt_chars = {
    "plain": {
        "ind": 2,
        "free": 1,
        "-": "-",
        "t": "+",
        "l": "+",
        "|": "|",
        ">": ">",
    },
    "fancy": {
        "ind": 2,
        "free": 1,
        "-": "─",
        "t": "├",
        "l": "└",
        "|": "│",
        ">": ">",
    },
}
fmt_chars["compact"] = merge_dicts(fmt_chars["plain"], {"free": 0})
fmt_chars["fancy_compact"] = merge_dicts(fmt_chars["fancy"], {"free": 0})


# helper to create a list of 3-tuples (target, depth, prefix) of an arbitrarily structured output
def _flatten_output(output, depth):
    if isinstance(output, (list, tuple, set)) or is_lazy_iterable(output):
        return [(outp, depth, "{}: ".format(i)) for i, outp in enumerate(output)]
    if isinstance(output, dict):
        return [(outp, depth, "{}: ".format(k)) for k, outp in six.iteritems(output)]
    return [(outp, depth, "") for outp in flatten(output)]


def _iter_output(output, offset, ind="  "):
    lookup = _flatten_output(output, 0)
    while lookup:
        output, odepth, oprefix = lookup.pop(0)
        ooffset = offset + odepth * ind

        if isinstance(output, Target):
            yield output, odepth, oprefix, ooffset, lookup

        else:
            # before updating the lookup list, but check if the output changes by this
            _lookup = _flatten_output(output, odepth + 1)
            if len(_lookup) > 0 and _lookup[0][0] == output:
                print(ooffset + oprefix + colored("not a target", color="red"))
            else:
                # print the key of the current structure
                print(ooffset + oprefix)

                # update the lookup list
                lookup[:0] = _lookup


def _print_wrapped(line, width, offset=""):
    # when the width is not set or the line is empty, just print the line
    if not line or width is None or width <= 0:
        print(line)
        return

    # split into actual strings to print (even parts) and color/style modifiers (odd parts) for
    # proper width computation
    parts = [(part, i % 2 == 1) for i, part in enumerate(uncolor_cre.split(line))]

    # build lines with odd parts until the line is filled
    line, length, last_style = "", 0, ""
    while parts:
        part, is_style = parts.pop(0)
        if is_style:
            # style modifier
            line += part
            last_style = part
        elif length + len(part) <= width:
            # actual string that still fits
            line += part
            length += len(part)
        else:
            # actual string that would overflow the line, so add the characters that would still fit
            # and then print the line
            n = width - length
            line += part[:n]
            print(line)
            # add the remaining characters with an uncolored offset and reset the state
            parts[:0] = [
                ("\033[0m", True),
                (uncolored(offset), False),
                (last_style, True),
                (part[n:], False),
            ]
            line, length, last_style = "", 0, ""
    # print any leftover line
    if line:
        print(line)


def print_task_deps(task, max_depth=1):
    max_depth = int(max_depth)

    print("print task dependencies with max_depth {}".format(max_depth))
    print("")

    # get the format chars
    cfg = Config.instance()
    fmt_name = cfg.get_expanded("task", "interactive_format")
    fmt = fmt_chars.get(fmt_name, fmt_chars["fancy"])

    # get the line break setting
    break_lines = cfg.get_expanded_bool("task", "interactive_line_breaks")
    out_width = cfg.get_expanded_int("task", "interactive_line_width")
    print_width = (out_width if out_width > 0 else get_terminal_width()) if break_lines else None
    _print = lambda line, offset: _print_wrapped(line, print_width, offset)

    parents_last_flags = []
    for dep, next_deps, depth, is_last in task.walk_deps(
        max_depth=max_depth,
        order="pre",
        yield_last_flag=True,
    ):
        del parents_last_flags[depth:]
        next_deps_shown = bool(next_deps) and (max_depth < 0 or depth < max_depth)

        # determine the print common offset
        offset = [(" " if f else fmt["|"]) + fmt["ind"] * " " for f in parents_last_flags[1:]]
        offset = "".join(offset)
        parents_last_flags.append(is_last)

        # print free space
        free_offset = offset + fmt["|"]
        free_lines = "\n".join(fmt["free"] * [free_offset])
        if depth > 0 and free_lines:
            print(free_lines)

        # determine task offset and prefix
        task_offset = offset
        if depth > 0:
            task_offset += fmt["l" if is_last else "t"] + fmt["ind"] * fmt["-"]
        task_prefix = "{} {} ".format(depth, fmt[">"])

        # determine text offset and prefix
        text_offset = offset
        if depth > 0:
            text_offset += (" " if is_last else fmt["|"]) + fmt["ind"] * " "
        text_prefix = (len(task_prefix) - 1) * " "
        text_offset += (fmt["|"] if next_deps_shown else " ") + text_prefix

        # print the task line
        _print(task_offset + task_prefix + dep.repr(color=True), text_offset)


def print_task_status(task, max_depth=0, target_depth=0, flags=None):
    from law.workflow.base import BaseWorkflow

    max_depth = int(max_depth)
    target_depth = int(target_depth)
    if flags:
        flags = tuple(flags.lower().split("-"))

    print("print task status with max_depth {} and target_depth {}".format(
        max_depth, target_depth))
    print("")

    # get the format chars
    cfg = Config.instance()
    fmt_name = cfg.get_expanded("task", "interactive_format")
    fmt = fmt_chars.get(fmt_name, fmt_chars["fancy"])

    # get the line break setting
    break_lines = cfg.get_expanded_bool("task", "interactive_line_breaks")
    out_width = cfg.get_expanded_int("task", "interactive_line_width")
    print_width = (out_width if out_width > 0 else get_terminal_width()) if break_lines else None
    _print = lambda line, offset: _print_wrapped(line, print_width, offset)

    # get other settings
    skip_seen = cfg.get_expanded_bool("task", "interactive_status_skip_seen")

    # walk through deps
    done = []
    parents_last_flags = []
    for dep, next_deps, depth, is_last in task.walk_deps(
        max_depth=max_depth,
        order="pre",
        yield_last_flag=True,
    ):
        del parents_last_flags[depth:]
        next_deps_shown = bool(next_deps) and (max_depth < 0 or depth < max_depth)

        # determine the print common offset
        offset = [(" " if f else fmt["|"]) + fmt["ind"] * " " for f in parents_last_flags[1:]]
        offset = "".join(offset)
        parents_last_flags.append(is_last)

        # print free space
        free_offset = offset + fmt["|"]
        free_lines = "\n".join(fmt["free"] * [free_offset])
        if depth > 0 and free_lines:
            print(free_lines)

        # when the dep is a workflow, independent of its create_branch_map_before_repr setting,
        # preload its branch map which updates branch parameters
        if isinstance(dep, BaseWorkflow):
            dep.get_branch_map()

        # determine task offset and prefix
        task_offset = offset
        if depth > 0:
            task_offset += fmt["l" if is_last else "t"] + fmt["ind"] * fmt["-"]
        task_prefix = "{} {} ".format(depth, fmt[">"])

        # determine text offset and prefix
        text_offset = offset
        if depth > 0:
            text_offset += (" " if is_last else fmt["|"]) + fmt["ind"] * " "
        text_prefix = (len(task_prefix) - 1) * " "
        text_offset += (fmt["|"] if next_deps_shown else " ") + text_prefix
        text_offset_ind = text_offset + fmt["ind"] * " "

        # print the task line
        _print(task_offset + task_prefix + dep.repr(color=True), text_offset)

        # skip if already seen
        if skip_seen and dep in done:
            _print(text_offset_ind + colored("outputs already checked", "yellow"), text_offset_ind)
            continue

        done.append(dep)

        # compiled regex for splitting leading whitespace
        ws_cre = re.compile(r"^(\s*)(.*)$")

        # start the traversing
        for output, _, oprefix, ooffset, _ in _iter_output(
            dep.output(),
            text_offset_ind,
            fmt["ind"] * " ",
        ):
            _print(ooffset + oprefix + output.repr(color=True), ooffset + len(oprefix) * " ")
            ooffset += fmt["ind"] * " "
            status_text = output.status_text(max_depth=target_depth, flags=flags, color=True)
            status_lines = status_text.split("\n")
            _print(ooffset + status_lines[0], ooffset)
            for line in status_lines[1:]:
                _print(ooffset + line, ooffset + ws_cre.match(line).group(1) + fmt["ind"] * " ")


def print_task_output(task, max_depth=0, scheme=True):
    max_depth = int(max_depth)
    scheme = flag_to_bool(scheme)

    print("print task output with max_depth {}, {} schemes\n".format(
        max_depth, "showing" if scheme else "hiding"))

    done = []
    for dep, _, depth in task.walk_deps(max_depth=max_depth, order="pre"):
        done.append(dep)

        for outp in flatten(dep.output()):
            kwargs = {}
            if isinstance(outp, (FileSystemTarget, FileCollection)):
                kwargs = {"scheme": scheme}
            for uri in make_list(outp.uri(**kwargs)):
                print(uri)


def remove_task_output(task, max_depth=0, mode=None, run_task=False):
    from law.task.base import ExternalTask
    from law.workflow.base import BaseWorkflow

    max_depth = int(max_depth)

    print("remove task output with max_depth {}".format(max_depth))

    run_task = flag_to_bool(run_task)
    if run_task:
        print("task will run after output removal")

    # get the format chars
    cfg = Config.instance()
    fmt_name = cfg.get_expanded("task", "interactive_format")
    fmt = fmt_chars.get(fmt_name, fmt_chars["fancy"])

    # get the line break setting
    break_lines = cfg.get_expanded_bool("task", "interactive_line_breaks")
    out_width = cfg.get_expanded_int("task", "interactive_line_width")
    print_width = [(out_width if out_width > 0 else get_terminal_width()) if break_lines else None]
    _print = lambda line, offset: _print_wrapped(line, print_width[0], offset)

    # custom query_choice function that updates the terminal_width
    def _query_choice(*args, **kwargs):
        if print_width[0]:
            print_width[0] = out_width if out_width > 0 else get_terminal_width()
        return query_choice(*args, **kwargs)

    # determine the mode, i.e., interactive, dry, all
    modes = ["i", "d", "a"]
    mode_names = ["interactive", "dry", "all"]
    if mode and mode not in modes:
        raise Exception("unknown removal mode '{}'".format(mode))
    if not mode:
        mode = _query_choice("removal mode?", modes, default="i", descriptions=mode_names)
    mode_name = mode_names[modes.index(mode)]
    print("selected {} mode".format(colored(mode_name, "blue", style="bright")))
    print("")

    done = []
    parents_last_flags = []
    for dep, next_deps, depth, is_last in task.walk_deps(
        max_depth=max_depth,
        order="pre",
        yield_last_flag=True,
    ):
        del parents_last_flags[depth:]
        next_deps_shown = bool(next_deps) and (max_depth < 0 or depth < max_depth)

        # determine the print common offset
        offset = [(" " if f else fmt["|"]) + fmt["ind"] * " " for f in parents_last_flags[1:]]
        offset = "".join(offset)
        parents_last_flags.append(is_last)

        # print free space
        free_offset = offset + fmt["|"]
        free_lines = "\n".join(fmt["free"] * [free_offset])
        if depth > 0 and free_lines:
            print(free_lines)

        # when the dep is a workflow, independent of its create_branch_map_before_repr setting,
        # preload its branch map which updates branch parameters
        if isinstance(dep, BaseWorkflow):
            dep.get_branch_map()

        # determine task offset and prefix
        task_offset = offset
        if depth > 0:
            task_offset += fmt["l" if is_last else "t"] + fmt["ind"] * fmt["-"]
        task_prefix = "{} {} ".format(depth, fmt[">"])

        # determine text offset and prefix
        text_offset = offset
        if depth > 0:
            text_offset += (" " if is_last else fmt["|"]) + fmt["ind"] * " "
        text_prefix = (len(task_prefix) - 1) * " "
        text_offset += (fmt["|"] if next_deps_shown else " ") + text_prefix
        text_offset_ind = text_offset + fmt["ind"] * " "

        # print the task line
        _print(task_offset + task_prefix + dep.repr(color=True), text_offset)

        # always skip external tasks
        if isinstance(dep, ExternalTask):
            _print(text_offset_ind + colored("task is external", "yellow"), text_offset_ind)
            continue

        # skip when this task was already handled
        if dep in done:
            _print(text_offset_ind + colored("already handled", "yellow"), text_offset_ind)
            continue
        done.append(dep)

        # skip when mode is "all" and task is configured to skip
        if mode == "a" and getattr(dep, "skip_output_removal", False):
            _print(text_offset_ind + colored("configured to skip", "yellow"), text_offset_ind)
            continue

        # query for a decision per task when mode is "interactive"
        task_mode = None
        if mode == "i":
            task_mode = _query_choice(text_offset_ind + "remove outputs?", ["y", "n", "a"],
                default="y", descriptions=["yes", "no", "all"])
            if task_mode == "n":
                continue

        # start the traversing through output structure
        for output, odepth, oprefix, ooffset, lookup in _iter_output(
            dep.output(),
            text_offset_ind,
            fmt["ind"] * " ",
        ):
            _print(ooffset + oprefix + output.repr(color=True), ooffset + len(oprefix) * " ")
            ooffset += fmt["ind"] * " "

            # skip external targets
            if getattr(output, "external", False):
                _print(ooffset + colored("external output", "yellow"), ooffset)
                continue

            # stop here when in dry mode
            if mode == "d":
                _print(ooffset + colored("dry removed", "yellow"), ooffset)
                continue

            # when the mode is "interactive" and the task decision is not "all", query per output
            if mode == "i" and task_mode != "a":
                if isinstance(output, TargetCollection):
                    coll_choice = _query_choice(ooffset + "remove?", ("y", "n", "i"),
                        default="n", descriptions=["yes", "no", "interactive"])
                    if coll_choice == "i":
                        lookup[:0] = _flatten_output(output.targets, odepth + 1)
                        continue
                    else:
                        target_choice = coll_choice
                else:
                    target_choice = _query_choice(ooffset + "remove?", ("y", "n"),
                        default="n", descriptions=["yes", "no"])
                if target_choice == "n":
                    _print(ooffset + colored("skipped", "yellow"), ooffset)
                    continue

            # finally remove
            output.remove()
            _print(ooffset + colored("removed", "red", style="bright"), ooffset)

    return run_task


def fetch_task_output(task, max_depth=0, mode=None, target_dir=".", include_external=False):
    from law.task.base import ExternalTask
    from law.workflow.base import BaseWorkflow

    max_depth = int(max_depth)
    print("fetch task output with max_depth {}".format(max_depth))

    target_dir = os.path.normpath(os.path.abspath(str(target_dir)))
    print("target directory is {}".format(target_dir))
    makedirs(target_dir)

    include_external = flag_to_bool(include_external)
    if include_external:
        print("include external tasks")

    # get the format chars
    cfg = Config.instance()
    fmt_name = cfg.get_expanded("task", "interactive_format")
    fmt = fmt_chars.get(fmt_name, fmt_chars["fancy"])

    # get the line break setting
    break_lines = cfg.get_expanded_bool("task", "interactive_line_breaks")
    out_width = cfg.get_expanded_int("task", "interactive_line_width")
    print_width = [(out_width if out_width > 0 else get_terminal_width()) if break_lines else None]
    _print = lambda line, offset: _print_wrapped(line, print_width[0], offset)

    # custom query_choice function that updates the terminal_width
    def _query_choice(*args, **kwargs):
        if print_width[0]:
            print_width[0] = out_width if out_width > 0 else get_terminal_width()
        return query_choice(*args, **kwargs)

    # determine the mode, i.e., all, dry, interactive
    modes = ["i", "a", "d"]
    mode_names = ["interactive", "all", "dry"]
    if mode is None:
        mode = _query_choice("fetch mode?", modes, default="i", descriptions=mode_names)
    elif isinstance(mode, int):
        mode = modes[mode]
    else:
        mode = mode[0].lower()
    if mode not in modes:
        raise Exception("unknown fetch mode '{}'".format(mode))
    mode_name = mode_names[modes.index(mode)]
    print("selected {} mode".format(colored(mode_name, "blue", style="bright")))
    print("")

    done = []
    parents_last_flags = []
    for dep, next_deps, depth, is_last in task.walk_deps(
        max_depth=max_depth,
        order="pre",
        yield_last_flag=True,
    ):
        del parents_last_flags[depth:]
        next_deps_shown = bool(next_deps) and (max_depth < 0 or depth < max_depth)

        # determine the print common offset
        offset = [(" " if f else fmt["|"]) + fmt["ind"] * " " for f in parents_last_flags[1:]]
        offset = "".join(offset)
        parents_last_flags.append(is_last)

        # print free space
        free_offset = offset + fmt["|"]
        free_lines = "\n".join(fmt["free"] * [free_offset])
        if depth > 0 and free_lines:
            print(free_lines)

        # when the dep is a workflow, independent of its create_branch_map_before_repr setting,
        # preload its branch map which updates branch parameters
        if isinstance(dep, BaseWorkflow):
            dep.get_branch_map()

        # determine task offset and prefix
        task_offset = offset
        if depth > 0:
            task_offset += fmt["l" if is_last else "t"] + fmt["ind"] * fmt["-"]
        task_prefix = "{} {} ".format(depth, fmt[">"])

        # determine text offset and prefix
        text_offset = offset
        if depth > 0:
            text_offset += (" " if is_last else fmt["|"]) + fmt["ind"] * " "
        text_prefix = (len(task_prefix) - 1) * " "
        text_offset += (fmt["|"] if next_deps_shown else " ") + text_prefix
        text_offset_ind = text_offset + fmt["ind"] * " "

        # print the task line
        _print(task_offset + task_prefix + dep.repr(color=True), text_offset)

        if not include_external and isinstance(dep, ExternalTask):
            _print(text_offset_ind + colored("task is external", "yellow"), text_offset_ind)
            continue

        if dep in done:
            _print(text_offset_ind + colored("outputs already fetched", "yellow"), text_offset_ind)
            continue

        if mode == "i":
            task_mode = _query_choice(text_offset_ind + "fetch outputs?", ("y", "n", "a"),
                default="y", descriptions=["yes", "no", "all"])
            if task_mode == "n":
                _print(text_offset_ind + colored("skipped", "yellow"), text_offset_ind)
                continue

        done.append(dep)

        # start the traversing through output structure with a lookup pattern
        for output, odepth, oprefix, ooffset, lookup in _iter_output(
            dep.output(),
            text_offset_ind,
            fmt["ind"] * " ",
        ):
            try:
                stat = output.stat()
            except:
                stat = None

            # print the target repr
            target_line = ooffset + oprefix + output.repr(color=True)
            if stat:
                target_line += " ({:.2f} {})".format(*human_bytes(stat.st_size))
            _print(target_line, ooffset + len(oprefix) * " ")
            ooffset += fmt["ind"] * " "

            # skip external targets
            if not include_external and getattr(output, "external", False):
                _print(ooffset + colored("external output, skip", "yellow"), ooffset)
                continue

            # skip missing targets
            if stat is None and not isinstance(output, TargetCollection):
                _print(ooffset + colored("not existing, skip", "yellow"), ooffset)
                continue

            # skip targets without a copy_to_local method
            is_copyable = callable(getattr(output, "copy_to_local", None))
            if not is_copyable and not isinstance(output, TargetCollection):
                _print(ooffset + colored("not a file target, skip", "yellow"), ooffset)
                continue

            # stop here when in dry mode
            if mode == "d":
                _print(ooffset + colored("dry fetched", "yellow"), ooffset)
                continue

            # collect actual outputs to fetch
            to_fetch = [output]
            if mode == "i" and task_mode != "a":
                if isinstance(output, TargetCollection):
                    coll_choice = _query_choice(ooffset + "fetch?", ("y", "n", "i"),
                        default="y", descriptions=["yes", "no", "interactive"])
                    if coll_choice == "i":
                        lookup[:0] = _flatten_output(output.targets, odepth + 1)
                        continue
                    else:
                        target_choice = coll_choice
                    to_fetch = list(output._flat_target_list)
                else:
                    target_choice = _query_choice(ooffset + "fetch?", ("y", "n"),
                        default="y", descriptions=["yes", "no"])
                if target_choice == "n":
                    _print(ooffset + colored("skipped", "yellow"), ooffset)
                    continue

            # flatten all target collections
            to_fetch_flat = []
            while to_fetch:
                t = to_fetch.pop(0)
                if isinstance(t, TargetCollection):
                    to_fetch[:0] = list(t._flat_target_list)
                else:
                    to_fetch_flat.append(t)

            # actual copy
            for outp in to_fetch_flat:
                if not callable(getattr(outp, "copy_to_local", None)):
                    continue

                basename = "{}__{}".format(dep.live_task_id, outp.basename)
                outp.copy_to_local(os.path.join(target_dir, basename), retries=0)

                _print(ooffset + "{} ({})".format(colored("fetched", "green", style="bright"),
                    basename), ooffset)
